from __future__ import annotations

import argparse
import json
import shutil
import subprocess
import sys
import tempfile
import webbrowser
from pathlib import Path

import nox
from nox import Session

from exasol.toolbox.nox._shared import DOCS_OUTPUT_DIR
from noxconfig import (
    PROJECT_CONFIG,
    Config,
)


def _build_docs(session: nox.Session, config: Config) -> None:
    session.run(
        "sphinx-build",
        "-W",
        "-b",
        "html",
        f"{config.doc}",
        DOCS_OUTPUT_DIR,
    )


def _build_multiversion_docs(session: nox.Session, config: Config) -> None:
    session.run(
        "sphinx-multiversion",
        f"{config.doc}",
        DOCS_OUTPUT_DIR,
    )
    session.run("touch", f"{DOCS_OUTPUT_DIR}/.nojekyll")


def _git_diff_changes_main() -> int:
    """
    Check if doc/changes is changed and return the exit code of command git diff.
    The exit code is 0 if there are no changes.
    """
    p = subprocess.run(
        [
            "git",
            "diff",
            "--quiet",
            "origin/main",
            "--",
            PROJECT_CONFIG.root / "doc/changes",
        ],
        capture_output=True,
    )
    return p.returncode


@nox.session(name="docs:multiversion", python=False)
def build_multiversion(session: Session) -> None:
    """Builds the multiversion project documentation"""
    _build_multiversion_docs(session, PROJECT_CONFIG)


@nox.session(name="docs:build", python=False)
def build_docs(session: Session) -> None:
    """Builds the project documentation"""
    _build_docs(session, PROJECT_CONFIG)


@nox.session(name="docs:open", python=False)
def open_docs(session: Session) -> None:
    """Opens the built project documentation"""
    docs_folder = PROJECT_CONFIG.root / DOCS_OUTPUT_DIR
    if not docs_folder.exists():
        session.error(f"No documentation could be found. {docs_folder} is missing")
    index = docs_folder / "index.html"
    webbrowser.open_new_tab(index.as_uri())


@nox.session(name="docs:clean", python=False)
def clean_docs(_session: Session) -> None:
    """Removes the documentations build folder"""
    docs_folder = PROJECT_CONFIG.root / DOCS_OUTPUT_DIR
    if docs_folder.exists():
        shutil.rmtree(docs_folder)


def _docs_list_links(doc_config: Path):
    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)
        sp = subprocess.run(  # nosec
            [
                "sphinx-build",
                "-b",
                "linkcheck",
                "-D",
                "linkcheck_ignore=.*",
                doc_config,
                tmpdir,
            ],
            capture_output=True,
            text=True,
        )
        if sp.returncode >= 2:
            return sp.returncode, sp.stderr
        output = tmpdir / "output.json"
        links = output.read_text().split("\n")
        file_links = []
        for link in links:
            if link != "":
                line = json.loads(link)
                if not line["uri"].startswith("#"):
                    file_links.append(line)
        file_links.sort(key=lambda file: file["filename"])
        return 0, "\n".join(
            f"filename: {fl['filename']}:{fl['lineno']} -> uri: {fl['uri']}"
            for fl in file_links
        )


def _docs_links_check(doc_config: Path, args):
    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)
        sp = subprocess.run(  # nosec
            [
                "sphinx-build",
                "-b",
                "linkcheck",
                doc_config,
                tmpdir,
            ],
        )
        if args.output and sp.returncode <= 1:
            result_json = tmpdir / "output.json"
            dst = Path(args.output) / "link-check-output.json"
            shutil.copyfile(result_json, dst)
            print(f"file generated at path: {result_json.resolve()}")
        return sp.returncode, (
            None if sp.returncode >= 2 else (tmpdir / "output.txt").read_text()
        )


@nox.session(name="links:list", python=False)
def docs_list_links(session: Session) -> None:
    """List all the links within the documentation."""
    r_code, text = _docs_list_links(PROJECT_CONFIG.doc)
    print(text)
    if r_code != 0:
        session.error()


@nox.session(name="links:check", python=False)
def docs_links_check(session: Session) -> None:
    """Checks whether all links in the documentation are accessible."""
    parser = argparse.ArgumentParser(
        prog="nox -s links:check",
        usage="nox -s links:check -- [-h] [-o |--output]",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-o", "--output", type=Path, help="path to copy the output json", default=None
    )
    args = parser.parse_args(session.posargs)
    r_code, problems = _docs_links_check(PROJECT_CONFIG.doc, args)
    if r_code >= 2:
        session.error(2)
    if r_code == 1 or problems != "":
        escape_red = "\033[31m"
        print(escape_red + "errors:")
        print(problems)
        session.error(1)


@nox.session(name="changelog:updated", python=False)
def updated(_session: Session) -> None:
    """Checks if the change log has been updated"""
    if _git_diff_changes_main() == 0:
        print(
            "Changelog is not updated.\n"
            "Please describe your changes in the changelog!"
        )
        sys.exit(1)



import subprocess  # nosec
from functools import wraps
from pathlib import Path


def run_command(func):
    @wraps(func)
    def wrapper(*args, **kwargs) -> str:
        command_list = func(*args, **kwargs)
        output = subprocess.run(
            command_list, capture_output=True, text=True, check=True
        )  # nosec
        return output.stdout.strip()

    return wrapper


class Git:
    @staticmethod
    @run_command
    def get_latest_tag():
        """
        Get the latest tag from the git repository.
        """
        return ["git", "describe", "--tags", "--abbrev=0"]

    @staticmethod
    @run_command
    def read_file_from_tag(tag: str, remote_file: str):
        """
        Read the contents of the specified file `remote_file` at the point in time
        specified by git tag `tag`.
        """
        return ["git", "cat-file", "blob", f"{tag}:{remote_file}"]

    @staticmethod
    def copy_remote_file_locally(
        tag: str, remote_file: str, destination_directory: Path
    ) -> None:
        """
        Copy the contents of the specified file `remote_file` at the point in time
        specified by git tag `tag` and copy it into the local `destination_directory/remote_file`.
        """
        contents = Git.read_file_from_tag(tag=tag, remote_file=remote_file)
        (destination_directory / remote_file).write_text(contents)

    @staticmethod
    @run_command
    def create_and_switch_to_branch(branch_name: str):
        return ["git", "switch", "-c", branch_name]



from __future__ import annotations

from collections.abc import Iterable

import nox
from nox import Session

from exasol.toolbox.nox._shared import (
    Mode,
    _version,
    python_files,
)
from noxconfig import (
    PROJECT_CONFIG,
    Config,
)

_PYUPGRADE_ARGS = ("--py39-plus",)


def _code_format(session: Session, mode: Mode, files: Iterable[str]) -> None:
    def command(*args: str) -> Iterable[str]:
        return args if mode == Mode.Fix else list(args) + ["--check"]

    session.run(*command("isort"), *files)
    session.run(*command("black"), *files)


def _pyupgrade(session: Session, config: Config, files: Iterable[str]) -> None:
    pyupgrade_args = getattr(config, "pyupgrade_args", _PYUPGRADE_ARGS)
    session.run(
        "pyupgrade",
        *pyupgrade_args,
        "--exit-zero-even-if-changed",
        *files,
    )


@nox.session(name="project:fix", python=False)
def fix(session: Session) -> None:
    """Runs all automated fixes on the code base"""
    py_files = python_files(PROJECT_CONFIG.root)
    _version(session, Mode.Fix)
    _pyupgrade(session, config=PROJECT_CONFIG, files=py_files)
    _code_format(session, Mode.Fix, py_files)


@nox.session(name="project:format", python=False)
def fmt_check(session: Session) -> None:
    """Checks the project for correct formatting"""
    py_files = python_files(PROJECT_CONFIG.root)
    _code_format(session=session, mode=Mode.Check, files=py_files)



from __future__ import annotations

__all__ = [
    "Mode",
    "build_docs",
    "check",
    "clean_docs",
    "coverage",
    "fix",
    "integration_tests",
    "lint",
    "open_docs",
    "prepare_release",
    "type_check",
    "unit_tests",
]


import nox
from nox import Session

from exasol.toolbox.nox._format import (
    _code_format,
    fix,
)

# fmt: off
# isort: off
from noxconfig import PROJECT_CONFIG


@nox.session(name="project:check", python=False)
def check(session: Session) -> None:
    """Runs all available checks on the project"""
    context = _context(session, coverage=True)
    py_files = python_files(PROJECT_CONFIG.root)
    _version(session, Mode.Check)
    _code_format(session, Mode.Check, py_files)
    _pylint(session, py_files)
    _type_check(session, py_files)
    _coverage(session, PROJECT_CONFIG, context)

from exasol.toolbox.nox._metrics import report
from exasol.toolbox.nox._test import (
    _coverage,
    coverage,
    integration_tests,
    unit_tests,
)
from exasol.toolbox.nox._lint import (
    _pylint,
    _type_check,
    lint,
    type_check,
)
from exasol.toolbox.nox._documentation import (
    build_docs,
    clean_docs,
    open_docs,
    updated,
)
from exasol.toolbox.nox._release import prepare_release
from exasol.toolbox.nox._shared import (
    Mode,
    _context,
    _version,
    python_files,
)

from exasol.toolbox.nox._ci import (
    python_matrix,
    exasol_matrix,
    full_matrix,
)

from exasol.toolbox.nox._release import prepare_release

from exasol.toolbox.nox._artifacts import (
    check_artifacts,
    copy_artifacts,
    upload_artifacts_to_sonar
)

from exasol.toolbox.nox._dependencies import (
    dependency_licenses,
    audit
)

from exasol.toolbox.nox._package_version import version_check

# isort: on
# fmt: on



from __future__ import annotations

import argparse
from dataclasses import dataclass
from pathlib import Path

import nox
from nox import Session

from exasol.toolbox.metrics import (
    Format,
    create_report,
    format_report,
)
from noxconfig import PROJECT_CONFIG


@dataclass
class RequiredFile:
    """
    Describes a required file and the related nox task to generate this
    file.
    """

    def __init__(self, file: Path | str, task: str):
        self.file = file if isinstance(file, Path) else PROJECT_CONFIG.root / file
        self.task = task

    def __str__(self) -> str:
        return f"{self.file.name} generated by `{self.task}`"


@nox.session(name="project:report", python=False)
def report(session: Session) -> None:
    """
    Collects and generates metrics summary for the workspace

    Attention:

        Pre-requisites:

        * Make sure you remove old and outdated artifacts
            - e.g. by running one of the following commands
                * :code:`git clean -xdf`
                * :code:`rm .coverage .lint.txt`

        * Run the following targets:
            - :code:`nox -s test:coverage`
            - :code:`nox -s lint:code`
            - :code:`nox -s lint:security`
    """
    formats = tuple(fmt.name.lower() for fmt in Format)
    usage = "nox -s project:report -- [options]"
    parser = argparse.ArgumentParser(
        description="Generates status report for the project", usage=usage
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default=formats[0],
        help="Output format to produce.",
        choices=formats,
    )
    required_files = [
        RequiredFile(".coverage", "test:coverage"),
        RequiredFile(".lint.txt", "lint:code"),
        RequiredFile(".security.json", "lint:security"),
    ]
    if missing_files := [f for f in required_files if not f.file.exists()]:
        missing = "\n- file ".join(str(f) for f in missing_files)
        session.error(
            "Some required files are missing.\n"
            "Please make sure you run the related nox tasks first:\n"
            f"{missing}"
        )
    sha1 = str(
        session.run("git", "rev-parse", "HEAD", external=True, silent=True)
    ).strip()
    args: argparse.Namespace = parser.parse_args(args=session.posargs)
    project_report = create_report(commit=sha1)
    fmt = Format.from_string(args.format)

    print(format_report(project_report, fmt))



import argparse
from argparse import (
    ArgumentParser,
    Namespace,
)
from inspect import cleandoc
from pathlib import Path

import nox
from nox import Session

from exasol.toolbox.util.version import Version
from noxconfig import (
    PROJECT_CONFIG,
    Config,
)

_SUCCESS = 0
_FAILURE = 1

# fmt: off
_VERSION_MODULE_TEMPLATE = cleandoc('''
    """
    ATTENTION:
        This file is generated by exasol/toolbox/nox/_package_version.py when using:
            * either "poetry run -- nox -s project:fix"
            * or "poetry run -- nox -s version:check -- --fix"
    Do not edit this file manually!
    If you need to change the version, do so in the pyproject.toml, e.g. by using
    `poetry version X.Y.Z`.
    """

    MAJOR = {major}
    MINOR = {minor}
    PATCH = {patch}
    VERSION = f"{{MAJOR}}.{{MINOR}}.{{PATCH}}"
    __version__ = VERSION
''') + "\n"
# fmt: on


def write_version_module(version: Version, version_file: Path) -> None:
    with version_file.open(mode="w", encoding="utf-8") as f:
        f.write(
            _VERSION_MODULE_TEMPLATE.format(
                major=version.major, minor=version.minor, patch=version.patch
            )
        )


def _create_parser() -> ArgumentParser:
    parser = ArgumentParser(
        prog="nox -s version:check --",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=False,
        help="enabled debug mode for execution.",
    )
    parser.add_argument(
        "-f",
        "--fix",
        action="store_true",
        default=False,
        help="updates the `version.py`, instead of performing a check.",
    )
    return parser


def _version_check(args: Namespace, config: Config) -> int:
    version_file = config.version_file

    module_version = Version.from_python_module(version_file)
    poetry_version = Version.from_poetry()
    if args.fix:
        print(
            f"Updating version in {version_file} from {module_version} to {poetry_version}"
        )
        write_version_module(version=poetry_version, version_file=version_file)
        module_version = Version.from_python_module(version_file)

    if module_version != poetry_version:
        print(
            f"Version in pyproject.toml ({poetry_version}) and {version_file} ({module_version}) do not match!"
        )
        return _FAILURE

    return _SUCCESS


@nox.session(name="version:check", python=False)
def version_check(session: Session) -> None:
    """
    Compare the version in the `version.py` to that
    declared in the `pyproject.toml`.
    """
    parser = _create_parser()
    args = parser.parse_args(session.posargs)
    if _version_check(args=args, config=PROJECT_CONFIG):
        session.error()



from __future__ import annotations

import argparse
import sys
from collections.abc import Iterable
from pathlib import Path

import nox
import rich.console
import tomlkit
from nox import Session

from exasol.toolbox.nox._shared import python_files
from noxconfig import PROJECT_CONFIG


def _pylint(session: Session, files: Iterable[str]) -> None:
    session.run(
        "pylint",
        "--output-format",
        "colorized,json:.lint.json,text:.lint.txt",
        *files,
    )


def _type_check(session: Session, files: Iterable[str]) -> None:
    session.run(
        "mypy",
        "--explicit-package-bases",
        "--namespace-packages",
        "--show-error-codes",
        "--pretty",
        "--show-column-numbers",
        "--show-error-context",
        "--scripts-are-modules",
        *files,
    )


def _security_lint(session: Session, files: Iterable[str]) -> None:
    session.run(
        "bandit",
        "--severity-level",
        "low",
        "--quiet",
        "--format",
        "json",
        "--output",
        ".security.json",
        "--exit-zero",
        *files,
    )
    session.run(
        "bandit",
        "--severity-level",
        "low",
        "--quiet",
        "--exit-zero",
        *files,
    )


def _import_lint(session: Session, path: Path) -> None:
    session.run("lint-imports", "--config", path)


class Dependencies:
    def __init__(self, illegal: dict[str, list[str]] | None):
        self._illegal = illegal or {}

    @staticmethod
    def parse(pyproject_toml: str) -> Dependencies:
        def _source_filter(version) -> bool:
            ILLEGAL_SPECIFIERS = ["url", "git", "path"]
            return any(specifier in version for specifier in ILLEGAL_SPECIFIERS)

        def find_illegal(part) -> list[str]:
            return [
                f"{name} = {version}"
                for name, version in part.items()
                if _source_filter(version)
            ]

        illegal: dict[str, list[str]] = {}
        toml = tomlkit.loads(pyproject_toml)
        poetry = toml.get("tool", {}).get("poetry", {})

        part = poetry.get("dependencies", {})
        if illegal_group := find_illegal(part):
            illegal["tool.poetry.dependencies"] = illegal_group

        part = poetry.get("dev", {}).get("dependencies", {})
        if illegal_group := find_illegal(part):
            illegal["tool.poetry.dev.dependencies"] = illegal_group

        part = poetry.get("group", {})
        for group, content in part.items():
            illegal_group = find_illegal(content.get("dependencies", {}))
            if illegal_group:
                illegal[f"tool.poetry.group.{group}.dependencies"] = illegal_group
        return Dependencies(illegal)

    @property
    def illegal(self) -> dict[str, list[str]]:
        return self._illegal


def report_illegal(illegal: dict[str, list[str]], console: rich.console.Console):
    count = sum(len(deps) for deps in illegal.values())
    suffix = "y" if count == 1 else "ies"
    console.print(f"{count} illegal dependenc{suffix}\n", style="red")
    for section, dependencies in illegal.items():
        console.print(f"\\[{section}]", style="red")
        for dependency in dependencies:
            console.print(dependency, style="red")
        console.print("")


@nox.session(name="lint:code", python=False)
def lint(session: Session) -> None:
    """Runs the static code analyzer on the project"""
    py_files = python_files(PROJECT_CONFIG.root / PROJECT_CONFIG.source)
    _pylint(session, py_files)


@nox.session(name="lint:typing", python=False)
def type_check(session: Session) -> None:
    """Runs the type checker on the project"""
    py_files = [f"{file}" for file in python_files(PROJECT_CONFIG.root)]
    _type_check(session, py_files)


@nox.session(name="lint:security", python=False)
def security_lint(session: Session) -> None:
    """Runs the security linter on the project"""
    py_files = python_files(PROJECT_CONFIG.root / PROJECT_CONFIG.source)
    _security_lint(session, py_files)


@nox.session(name="lint:dependencies", python=False)
def dependency_check(session: Session) -> None:
    """Checks if only valid sources of dependencies are used"""
    content = Path(PROJECT_CONFIG.root, "pyproject.toml").read_text()
    dependencies = Dependencies.parse(content)
    console = rich.console.Console()
    if illegal := dependencies.illegal:
        report_illegal(illegal, console)
        sys.exit(1)


@nox.session(name="lint:import", python=False)
def import_lint(session: Session) -> None:
    """(experimental) Runs import linter on the project"""
    parser = argparse.ArgumentParser(
        usage="nox -s import-lint -- [options]",
        description="Runs the import linter on the project",
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="path to the configuration file for the importlinter",
        metavar="TEXT",
    )

    args: argparse.Namespace = parser.parse_args(args=session.posargs)
    file: str = args.config
    path: Path | None = None
    if file is None:
        path = getattr(
            PROJECT_CONFIG, "import_linter_config", Path(".import_linter_config")
        )
    else:
        path = Path(file)
    if not path.exists():
        session.error(
            "Please make sure you have a configuration file for the importlinter"
        )
    _import_lint(session=session, path=path)



from __future__ import annotations

import argparse
import re
import subprocess
from pathlib import Path

import nox
from nox import Session

from exasol.toolbox.nox._shared import (
    Mode,
    _version,
)
from exasol.toolbox.nox.plugin import NoxTasks
from exasol.toolbox.util.git import Git
from exasol.toolbox.util.release.changelog import Changelogs
from exasol.toolbox.util.version import (
    ReleaseTypes,
    Version,
)
from noxconfig import PROJECT_CONFIG


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="nox -s release:prepare",
        usage="nox -s release:prepare -- [-h] [-t | --type {major,minor,patch}]",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-t",
        "--type",
        type=ReleaseTypes,
        help="specifies which type of upgrade is to be performed",
        required=True,
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--no-add",
        default=False,
        action="store_true",
        help="Neither add nor commit the changes",
    )
    parser.add_argument(
        "--no-branch",
        default=False,
        action="store_true",
        help="Do not create a branch to commit the changes on",
    )
    parser.add_argument(
        "--no-pr",
        default=False,
        action="store_true",
        help="Do not create a pull request for the changes",
    )
    return parser


def _update_project_version(session: Session, version: Version) -> Version:
    session.run("poetry", "version", f"{version}")
    _version(session, Mode.Fix)
    return version


def _add_files_to_index(session: Session, files: list[Path]) -> None:
    for file in files:
        session.run("git", "add", f"{file}")


class ReleaseError(Exception):
    """Error during trigger release"""


def _trigger_release(project_config) -> Version:
    def run(*args: str):
        try:
            return subprocess.run(
                args, capture_output=True, text=True, check=True
            ).stdout
        except subprocess.CalledProcessError as ex:
            raise ReleaseError(
                f"failed to execute command {ex.cmd}\n\n{ex.stderr}"
            ) from ex

    branches = run("git", "remote", "show", "origin")
    if not (result := re.search(r"HEAD branch: (\S+)", branches)):
        raise ReleaseError("default branch could not be found")
    default_branch = result.group(1)

    run("git", "checkout", default_branch)
    run("git", "pull")

    release_version: Version = Version.from_poetry()
    print(f"release version: {release_version}")

    if re.search(rf"{release_version}", run("git", "tag", "--list")):
        raise ReleaseError(f"tag {release_version} already exists")
    if re.search(rf"{release_version}", run("gh", "release", "list")):
        raise ReleaseError(f"release {release_version} already exists")

    run("git", "tag", str(release_version))
    run("git", "push", "origin", str(release_version))

    if (
        hasattr(project_config, "create_major_version_tags")
        and project_config.create_major_version_tags
    ):
        major_release_version = f"v{release_version.major}"
        run("git", "tag", "-f", str(major_release_version))
        run("git", "push", "-f", "origin", str(major_release_version))

    return release_version


@nox.session(name="release:prepare", python=False)
def prepare_release(session: Session) -> None:
    """
    Prepare the project for a new release.
    """
    parser = _create_parser()
    args = parser.parse_args(session.posargs)
    new_version = Version.upgrade_version_from_poetry(args.type)

    if not args.no_branch and not args.no_add:
        Git.create_and_switch_to_branch(f"release/prepare-{new_version}")

    _ = _update_project_version(session, new_version)

    changelogs = Changelogs(
        changes_path=PROJECT_CONFIG.doc / "changes",
        root_path=PROJECT_CONFIG.root,
        version=new_version,
    )
    changelogs.update_changelogs_for_release()
    changed_files = changelogs.get_changed_files()

    pm = NoxTasks.plugin_manager(PROJECT_CONFIG)
    pm.hook.prepare_release_update_version(
        session=session, config=PROJECT_CONFIG, version=new_version
    )

    if args.no_add:
        return

    changed_files += [
        PROJECT_CONFIG.root / "pyproject.toml",
        PROJECT_CONFIG.version_file,
    ]
    results = pm.hook.prepare_release_add_files(session=session, config=PROJECT_CONFIG)
    changed_files += [f for plugin_response in results for f in plugin_response]
    _add_files_to_index(
        session,
        changed_files,
    )
    session.run("git", "commit", "-m", f"Prepare release {new_version}")

    if not args.no_pr:
        session.run(
            "gh",
            "pr",
            "create",
            "--title",
            f"Prepare release {new_version}",
            "--body",
            '""',
        )


@nox.session(name="release:trigger", python=False)
def trigger_release(session: Session) -> None:
    """trigger an automatic project release"""
    print(f"new version: {_trigger_release(PROJECT_CONFIG)}")



import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import (
    Callable,
    Optional,
)

from exasol.toolbox.util.version import Version


def update_github_yml(template: Path, version: Version) -> None:
    """Updates versions in GitHub workflows and actions"""
    with open(template, encoding="utf-8") as file:
        content = file.readlines()

    content = update_versions(lines=content, version=version)

    with open(template, "w", encoding="utf-8") as file:
        file.writelines(content)


@dataclass(frozen=True)
class Pattern:
    match_pattern: str
    break_pattern: str
    version_pattern: str

    @property
    def full_pattern(self) -> str:
        return f"{self.match_pattern}{self.break_pattern}{self.version_pattern}"

    def replace_version(self, line: str, version: str) -> str:
        return re.sub(
            f"{self.break_pattern}{self.version_pattern}",
            f"{self.break_pattern}{version}",
            line,
        )


def full_version_modifier(version: Version) -> str:
    return str(version)


def major_version_modifier(version: Version) -> str:
    return f"v{version.major}"


class GithubActionsReplacement:
    def __init__(
        self, pattern: Pattern, version_string_modifier: Callable[[Version], str]
    ) -> None:
        self.pattern = pattern
        self.version_string_modifier = version_string_modifier

    def replace_version(self, line: str, version: Version) -> Optional[str]:
        match = re.search(self.pattern.full_pattern, line)
        if match:
            return self.pattern.replace_version(
                line=line, version=self.version_string_modifier(version)
            )
        return None


class Replacements(Enum):
    github = GithubActionsReplacement(
        pattern=Pattern(
            match_pattern="exasol/python-toolbox/.github/[^/]+/[^/]+",
            break_pattern="@",
            version_pattern=r"v[0-9]+",
        ),
        version_string_modifier=major_version_modifier,
    )

    pypi = GithubActionsReplacement(
        pattern=Pattern(
            match_pattern="exasol-toolbox",
            break_pattern="==",
            version_pattern=r"[0-9]+\.[0-9]+\.[0-9]+",
        ),
        version_string_modifier=full_version_modifier,
    )


def _update_line_with_version(line: str, version: Version) -> str:
    for replacement in Replacements:
        if replaced_line := replacement.value.replace_version(
            line=line, version=version
        ):
            return replaced_line
    return line


def update_versions(lines: list[str], version: Version) -> list[str]:
    return [_update_line_with_version(line=line, version=version) for line in lines]



"""This module contains security related CLI tools and code"""

from __future__ import annotations

import json
import re
import subprocess
import sys
from collections.abc import (
    Generator,
    Iterable,
)
from dataclasses import (
    asdict,
    dataclass,
)
from enum import Enum
from functools import partial
from inspect import cleandoc
from pathlib import Path
from typing import Optional

import typer

stdout = print
stderr = partial(print, file=sys.stderr)


# Note: In the long term we may want to adapt the official CVE json schema
# https://github.com/CVEProject/cve-schema/blob/master/schema/v5.0/CVE_JSON_5.0_schema.json
@dataclass(frozen=True)
class Issue:
    cve: str
    cwe: str
    description: str
    coordinates: str
    references: tuple


def _issues(input) -> Generator[Issue]:
    lines = (l for l in input if l.strip() != "")
    for line in lines:
        obj = json.loads(line)
        yield Issue(**obj)


def _issues_as_json_str(issues):
    for issue in issues:
        issue = asdict(issue)
        yield json.dumps(issue)


def gh_security_issues() -> Generator[tuple[str, str]]:
    """
    Yields issue-id, cve-id pairs for all (closed, open) issues associated with CVEs

    Return:
        A generator which yields tuples of (id,title).

    Raises:
        subprocess.CalledProcessError: If the underlying command fails.
    """
    # fmt: off
    command = [
        "gh", "issue", "list",
        "--label", "security",
        "--search", "CVE",
        "--json", "id,title",
        "--limit", "1000",
    ]
    # fmt: on
    try:
        result = subprocess.run(command, check=True, capture_output=True)
    except FileNotFoundError as ex:
        msg = "Command 'gh' not found. Please make sure you have installed the github cli."
        raise FileNotFoundError(msg) from ex
    except subprocess.CalledProcessError as ex:
        stderr(f"{ex}")
        raise ex

    cve_pattern = re.compile(r"CVE-\d{4}-\d{4,7}")
    issues = json.loads(result.stdout.decode("utf-8"))
    issues = (
        (issue["id"], cve_pattern.search(issue["title"]).group())  # type: ignore
        for issue in issues
        if cve_pattern.search(issue["title"])
    )
    return issues


def from_maven(report: str) -> Iterable[Issue]:
    # Note: Consider adding warnings if there is the same cve with multiple coordinates
    report = json.loads(report)
    dependencies = report.get("vulnerable", {})  # type: ignore
    for dependency_name, dependency in dependencies.items():  # type: ignore
        for v in dependency["vulnerabilities"]:  # type: ignore
            references = [v["reference"]] + v["externalReferences"]
            yield Issue(
                cve=v["cve"],
                cwe=v["cwe"],
                description=v["description"],
                coordinates=dependency_name,
                references=tuple(references),
            )


class VulnerabilitySource(str, Enum):
    CVE = "CVE"
    CWE = "CWE"
    GHSA = "GHSA"
    PYSEC = "PYSEC"

    @classmethod
    def from_prefix(cls, name: str) -> VulnerabilitySource | None:
        for el in cls:
            if name.upper().startswith(el.value):
                return el
        return None

    def get_link(self, package: str, vuln_id: str) -> str:
        if self == VulnerabilitySource.CWE:
            cwe_id = vuln_id.upper().replace(f"{VulnerabilitySource.CWE.value}-", "")
            return f"https://cwe.mitre.org/data/definitions/{cwe_id}.html"

        map_link = {
            VulnerabilitySource.CVE: "https://nvd.nist.gov/vuln/detail/{vuln_id}",
            VulnerabilitySource.GHSA: "https://github.com/advisories/{vuln_id}",
            VulnerabilitySource.PYSEC: "https://github.com/pypa/advisory-database/blob/main/vulns/{package}/{vuln_id}.yaml",
        }
        return map_link[self].format(package=package, vuln_id=vuln_id)


def identify_pypi_references(
    references: list[str], package_name: str
) -> tuple[list[str], list[str], list[str]]:
    refs: dict = {k: [] for k in VulnerabilitySource}
    links = []
    for reference in references:
        if source := VulnerabilitySource.from_prefix(reference.upper()):
            refs[source].append(reference)
            links.append(source.get_link(package=package_name, vuln_id=reference))
    return (
        refs[VulnerabilitySource.CVE],
        refs[VulnerabilitySource.CWE],
        links,
    )


def from_pip_audit(report: str) -> Iterable[Issue]:
    """
    Transforms the JSON output from `nox -s dependency:audit` into an iterable of
    `security.Issue` objects.

    This does not gracefully handle scenarios where:
     - a CVE is not initially associated with the vulnerability; however, the assumption
     is that such vulnerabilities will later be associated with a CVE.
     - the same vulnerability ID (CVE, PYSEC, GHSA, etc.) is present across
     multiple coordinates.

    Input as string:
        [
          {
            "name": "jinja2",
            "version": "3.1.5",
            "refs": [
              "GHSA-cpwx-vrp4-4pq7",
              "CVE-2025-27516"
            ],
            "description": "An oversight ..."
          }
        ]


    Args:
        report:
            the JSON output of `nox -s dependency:audit` provided as a str
    """
    vulnerabilities = json.loads(report)

    for vulnerability in vulnerabilities:
        cves, cwes, links = identify_pypi_references(
            references=vulnerability["refs"], package_name=vulnerability["name"]
        )
        if cves:
            yield Issue(
                cve=sorted(cves)[0],
                cwe="None" if not cwes else ", ".join(cwes),
                description=vulnerability["description"],
                coordinates=f"{vulnerability['name']}:{vulnerability['version']}",
                references=tuple(links),
            )


@dataclass(frozen=True)
class SecurityIssue:
    file_name: str
    line: int
    column: int
    cwe: str
    test_id: str
    description: str
    references: tuple


def from_json(report_str: str, prefix: Path) -> Iterable[SecurityIssue]:
    report = json.loads(report_str)
    issues = report.get("results", {})
    for issue in issues:
        references = []
        if issue["more_info"]:
            references.append(issue["more_info"])
        if issue.get("issue_cwe", {}).get("link", None):
            references.append(issue["issue_cwe"]["link"])
        yield SecurityIssue(
            file_name=issue["filename"].replace(str(prefix) + "/", ""),
            line=issue["line_number"],
            column=issue["col_offset"],
            cwe=str(issue["issue_cwe"].get("id", "")),
            test_id=issue["test_id"],
            description=issue["issue_text"],
            references=tuple(references),
        )


def issues_to_markdown(issues: Iterable[SecurityIssue]) -> str:
    template = cleandoc(
        """
        {header}{rows}
    """
    )

    def _header():
        header = "# Security\n\n"
        header += "|File|line/<br>column|Cwe|Test ID|Details|\n"
        header += "|---|:-:|:-:|:-:|---|\n"
        return header

    def _row(issue):
        row = "|" + issue.file_name + "|"
        row += f"line: {issue.line}<br>column: {issue.column}|"
        row += issue.cwe + "|"
        row += issue.test_id + "|"
        for element in issue.references:
            row += element + " ,<br>"
        row = row[:-5] + "|"
        return row

    return template.format(header=_header(), rows="\n".join(_row(i) for i in issues))


def security_issue_title(issue: Issue) -> str:
    return f"{issue.cve}: {issue.coordinates}"


def security_issue_body(issue: Issue) -> str:
    def as_markdown_listing(elements: Iterable[str]):
        return "\n".join(f"- {element}" for element in elements)

    body = cleandoc(
        """
        ## Summary
        {description}

        CVE: {cve}
        CWE: {cwe}

        ## References
        {references}
        """
    )
    return body.format(
        cve=issue.cve,
        cwe=issue.cwe,
        description=issue.description,
        references=as_markdown_listing(issue.references),
    )


def create_security_issue(
    issue: Issue, project: Optional[str] = None
) -> tuple[str, str]:
    # fmt: off
    command = [
        "gh", "issue", "create",
        "--label", "security",
        "--title", security_issue_title(issue),
        "--body", security_issue_body(issue),
    ]
    if project:
        command.extend(['--project', project])
    # fmt: on
    try:
        result = subprocess.run(command, check=True, capture_output=True)
    except FileNotFoundError as ex:
        msg = "Command 'gh' not found. Please make sure you have installed the github cli."
        raise FileNotFoundError(msg) from ex
    except subprocess.CalledProcessError as ex:
        stderr(f"{ex}")
        raise ex

    std_err = result.stderr.decode("utf-8")
    std_out = result.stdout.decode("utf-8")

    return std_err, std_out


CLI = typer.Typer()
CVE_CLI = typer.Typer()
CLI.add_typer(CVE_CLI, name="cve", help="Work with CVE's")


class Format(str, Enum):
    Maven = "maven"
    PipAudit = "pip-audit"


# pylint: disable=redefined-builtin
@CVE_CLI.command(name="convert")
def convert(
    format: Format = typer.Argument(..., help="input format to be converted."),
    input_file: typer.FileText = typer.Argument(
        default="-", mode="r", help="input file which shall be converted"
    ),
) -> None:
    """
    Convert a language or tool specific security report into a list of cve's in the jsonl format

    Output:
    { "cve": "<cve-id>", "cwe": "<cwe-id>", "description": "<multiline string>", "coordinates": "<string>", "references": ["<url>", "<url>", ...] }
    """

    def _maven(infile):
        issues = from_maven(infile.read())
        for issue in _issues_as_json_str(issues):
            stdout(issue)
        raise typer.Exit(code=0)

    def _pip_audit(infile):
        issues = from_pip_audit(infile.read())
        for issue in _issues_as_json_str(issues):
            stdout(issue)
        raise typer.Exit(code=0)

    actions = {Format.Maven: _maven, Format.PipAudit: _pip_audit}
    action = actions[format]
    action(input_file)


class Filter(str, Enum):
    GitHubIssues = "github-issues"
    PassThrough = "pass-through"


# pylint: disable=redefined-builtin
@CVE_CLI.command(name="filter")
def filter(
    type: Filter = typer.Argument(help="filter type to apply"),
    input_file: typer.FileText = typer.Argument(
        default="-", mode="r", help="file containing cve's in the jsonl format"
    ),
) -> None:
    """
    Filter specific CVE's from the input


    Input:
    { "cve": "<cve-id>", "cwe": "<cwe-id>", "description": "<multiline string>", "coordinates": "<string>", "references": ["<url>", "<url>", ...] }
    """

    def _github(infile):
        to_be_filtered = {cve for _, cve in gh_security_issues()}
        stderr("Filtering:")
        for issue in to_be_filtered:
            stderr(f"{issue}")
        filtered_issues = [
            issue for issue in _issues(infile) if issue.cve not in to_be_filtered
        ]
        for issue in _issues_as_json_str(filtered_issues):
            stdout(issue)
        raise typer.Exit(code=0)

    def _pass_through(infile):
        for line in infile:
            stdout(line)
        raise typer.Exit(code=0)

    actions = {Filter.GitHubIssues: _github, Filter.PassThrough: _pass_through}
    action = actions[type]
    action(input_file)


@CVE_CLI.command(name="create")
def create(
    input_file: typer.FileText = typer.Argument(
        default="-", mode="r", help="file of cve's in the jsonl format"
    ),
    project: str = typer.Option(
        default="", help="Project the created ticket shall be associated with."
    ),
) -> None:
    """
    Create GitHub issues for CVE's

    Input:
    { "cve": "<cve-id>", "cwe": "<cwe-id>", "description": "<multiline string>", "coordinates": "<string>", "references": ["<url>", "<url>", ...] }

    Output:
    Links to the created issue(s)
    """
    for issue in _issues(input_file):
        std_err, issue_url = create_security_issue(issue, project)
        stderr(std_err)
        stdout(format_jsonl(issue_url, issue))


class PPrintFormats(str, Enum):
    markdown = "markdown"


@CLI.command(name="pretty-print")
def json_issue_to_markdown(
    json_file: typer.FileText = typer.Argument(
        mode="r", help="json file with issues to convert"
    ),
    path: Path = typer.Argument(default=Path("."), help="path to project root"),
) -> None:
    content = json_file.read()
    issues = from_json(content, path.absolute())
    issues = sorted(issues, key=lambda i: (i.file_name, i.cwe, i.test_id))
    print(issues_to_markdown(issues))


def format_jsonl(issue_url: str, issue: Issue) -> str:
    issue_json = asdict(issue)
    issue_json["issue_url"] = issue_url.strip()
    return json.dumps(issue_json)


if __name__ == "__main__":
    CLI()



"""
Git based multi version support for sphinx.

This module/plugin provides support for creating a multi version output for the documented project.
Version's are defined using git tags and branches, the only supported output format is html.

NOTICE:
The original version of this package was published under the BSD license and can be found `here <https://github.com/Holzhaus/sphinx-multiversion>`_.
We also took some patches from `this <https://github.com/samuel-emrys/sphinx-multiversion>`_ which was also published under the BSD license.

So huge thanks to the original author `Jan Holthuis <https://github.com/Holzhaus>`_ and all contributors.


NOTE:
The original version and its defaults were minimally adjusted to work in Exasol's projects with the defaults and without adding an extra template. The [Shibuya](https://github.com/lepture/shibuya) theme is expected to be used for HTML, which already evaluates the versions field in the HTML context and generates an appropriate selector for the versions.
"""

from exasol.toolbox.sphinx.multiversion.main import main
from exasol.toolbox.sphinx.multiversion.sphinx import setup
from exasol.toolbox.version import VERSION

__version__ = VERSION

__all__ = [
    "setup",
    "main",
]



from __future__ import annotations

import subprocess
from dataclasses import dataclass
from enum import Enum
from functools import (
    total_ordering,
    wraps,
)
from pathlib import Path
from shutil import which
from typing import Any

from exasol.toolbox.error import ToolboxError


def _index_or(container, index, default):
    try:
        return container[index]
    except IndexError:
        return default


class ReleaseTypes(Enum):
    Major = "major"
    Minor = "minor"
    Patch = "patch"

    def __str__(self):
        return self.name.lower()


def poetry_command(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        cmd = which("poetry")
        if not cmd:
            raise ToolboxError("Couldn't find poetry executable")
        try:
            return func(*args, **kwargs)
        except subprocess.CalledProcessError as ex:
            raise ToolboxError(f"Failed to execute: {ex.cmd}") from ex

    return wrapper


@total_ordering
@dataclass(frozen=True)
class Version:
    major: int
    minor: int
    patch: int

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"

    def __lt__(self, other: object):
        if not isinstance(other, Version):
            return NotImplemented
        return (
            self.major < other.major
            or (self.major <= other.major and self.minor < other.minor)
            or (
                self.major <= other.major
                and self.minor <= other.minor
                and self.patch < other.patch
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, Version):
            return NotImplemented
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    @staticmethod
    def from_string(version):
        parts = [int(number, base=0) for number in version.split(".")]
        if len(parts) > 3:
            raise ValueError(
                "Version has an invalid format, "
                f"expected: '<major>.<minor>.<patch>', actual: '{version}'"
            )
        version = [_index_or(parts, i, 0) for i in range(3)]
        return Version(*version)

    @staticmethod
    @poetry_command
    def from_poetry():
        output = subprocess.run(
            ["poetry", "version", "--no-ansi", "--short"],
            capture_output=True,
            text=True,
        )
        return Version.from_string(output.stdout.strip())

    @staticmethod
    @poetry_command
    def upgrade_version_from_poetry(t: ReleaseTypes):
        output = subprocess.run(
            ["poetry", "version", str(t), "--dry-run", "--no-ansi", "--short"],
            capture_output=True,
            text=True,
        )
        return Version.from_string(output.stdout.strip())

    @staticmethod
    def from_python_module(path: Path) -> Version:
        """Retrieve version information from the `version` module"""
        with open(path, encoding="utf-8") as file:
            _locals: dict[str, Any] = {}
            _globals: dict[str, Any] = {}
            exec(file.read(), _locals, _globals)

            try:
                version = _globals["VERSION"]
            except KeyError as ex:
                raise ToolboxError("Couldn't find version within module") from ex

            return Version.from_string(version)



import typer

from exasol.toolbox.tools import (
    issue,
    lint,
    security,
    workflow,
)

CLI = typer.Typer()
CLI.add_typer(workflow.CLI, name="workflow", help="Manage github workflows")
CLI.add_typer(security.CLI, name="security", help="Security related helpers")
CLI.add_typer(issue.CLI, name="issue", help="Manage issue templates")
CLI.add_typer(lint.CLI, name="lint", help="linting related helpers")

if __name__ == "__main__":
    CLI()



import json
from collections.abc import Iterable
from dataclasses import dataclass
from inspect import cleandoc
from pathlib import Path

import typer

CLI = typer.Typer()


@dataclass(frozen=True)
class Finding:
    type: str
    module: str
    obj: str
    line: int
    column: int
    endLine: int
    endColumn: int
    path: str
    symbol: str
    message: str
    message_id: str


def lint_issue_from_json(data: str) -> Iterable[Finding]:
    issues = json.loads(data)
    for issue in issues:
        yield Finding(
            type=issue["type"],
            module=issue["module"],
            obj=issue["obj"],
            line=issue["line"],
            column=issue["column"],
            endLine=issue["endLine"],
            endColumn=issue["endColumn"],
            path=issue["path"],
            symbol=issue["symbol"],
            message=issue["message"],
            message_id=issue["message-id"],
        )


def lint_issue_to_markdown(lint_issues: Iterable[Finding]) -> str:
    def _header() -> str:
        header = "# Static Code Analysis\n\n"
        header += "|File|line/<br>column|id|message|\n"
        header += "|---|:-:|:-:|---|\n"
        return header

    def _rows(findings: Iterable[Finding]) -> str:
        rows = ""
        for finding in findings:
            rows += f"|{finding.path}"
            rows += f"|line: {finding.line}/<br>column: {finding.column}"
            rows += f"|{finding.message_id}"
            rows += f"|{finding.message}|\n"
        return rows

    template = cleandoc(
        """
        {header}{rows}
        """
    )
    lint_issues = sorted(lint_issues, key=lambda i: (i.path, i.message_id, i.line))
    return template.format(header=_header(), rows=_rows(lint_issues))


@CLI.command(name="pretty-print")
def lint_json_to_markdown(
    path: Path = typer.Argument(default=Path(".lint.json"), help="path to lint.json"),
) -> None:
    """converts the lint json to a Markdown table"""
    issues = lint_issue_from_json(path.read_text())
    print(lint_issue_to_markdown(issues))


if __name__ == "__main__":
    CLI()



from __future__ import annotations

from collections.abc import (
    Iterable,
    MutableMapping,
)
from pathlib import Path
from typing import Any

import nox
from nox import Session

from exasol.toolbox.nox._shared import _context
from exasol.toolbox.nox.plugin import NoxTasks
from noxconfig import (
    PROJECT_CONFIG,
    Config,
)


def _test_command(
    path: Path, config: Config, context: MutableMapping[str, Any]
) -> Iterable[str]:
    coverage_command = (
        ["coverage", "run", "-a", f"--rcfile={config.root / 'pyproject.toml'}", "-m"]
        if context["coverage"]
        else []
    )
    pytest_command = ["pytest", "-v", f"{path}"]
    return coverage_command + pytest_command + context["fwd-args"]


def _unit_tests(
    session: Session, config: Config, context: MutableMapping[str, Any]
) -> None:
    command = _test_command(config.root / "test" / "unit", config, context)
    session.run(*command)


def _integration_tests(
    session: Session, config: Config, context: MutableMapping[str, Any]
) -> None:
    pm = NoxTasks.plugin_manager(config)

    # run pre integration test plugins
    pm.hook.pre_integration_tests_hook(session=session, config=config, context=context)

    # Notes:
    # - Catch exceptions and ensure post-hooks run before exiting
    # - Consider making the executed command(s) configurable via a plugin hook
    #   (The default implementation of the hook could provide the current implementation)
    command = _test_command(config.root / "test" / "integration", config, context)
    session.run(*command)

    # run post integration test plugins
    pm.hook.post_integration_tests_hook(session=session, config=config, context=context)


def _pass(
    _session: Session, _config: Config, _context: MutableMapping[str, Any]
) -> bool:
    """No operation"""
    return True


def _coverage(
    session: Session, config: Config, context: MutableMapping[str, Any]
) -> None:
    command = ["coverage", "report", "-m"]
    coverage_file = config.root / ".coverage"
    coverage_file.unlink(missing_ok=True)
    _unit_tests(session, config, context)
    _integration_tests(session, config, context)
    session.run(*command)


@nox.session(name="test:unit", python=False)
def unit_tests(session: Session) -> None:
    """Runs all unit tests"""
    context = _context(session)
    _unit_tests(session, PROJECT_CONFIG, context)


@nox.session(name="test:integration", python=False)
def integration_tests(session: Session) -> None:
    """
    Runs the all integration tests

    If a project needs to execute code pre-/post the test execution,
    it should provide appropriate hooks on their config object.
        * pre_integration_tests_hook(session: Session, config: Config, context: MutableMapping[str, Any]) -> bool:
        * post_integration_tests_hook(session: Session, config: Config, context: MutableMapping[str, Any]) -> bool:
    """
    context = _context(session)
    _integration_tests(session, PROJECT_CONFIG, context)


@nox.session(name="test:coverage", python=False)
def coverage(session: Session) -> None:
    """Runs all tests (unit + integration) and reports the code coverage"""
    context = _context(session, coverage=True)
    _coverage(session, PROJECT_CONFIG, context)



from pathlib import Path

import typer

from exasol.toolbox.tools import template

CLI = typer.Typer()
PKG = "exasol.toolbox.templates.github.workflows"
TEMPLATE_TYPE = "workflow"
LEXER = "yaml"


@CLI.command(name="list")
def list_workflows(
    columns: bool = typer.Option(
        False, "--columns", "-c", help="use column style presentation like `ls`"
    )
) -> None:
    """List all available workflows."""
    template.list_templates(columns=columns, pkg=PKG)


@CLI.command(name="show")
def show_workflow(
    workflow: str = typer.Argument(..., help="Workflow which shall be shown."),
) -> None:
    """Shows a specific workflow."""
    template.show_templates(
        template=workflow, pkg=PKG, template_type=TEMPLATE_TYPE, lexer=LEXER
    )


@CLI.command(name="diff")
def diff_workflow(
    workflow: str = typer.Argument(..., help="workflow which shall be diffed."),
    dest: Path = typer.Argument(
        Path("./.github/workflows"),
        help="target directory to diff the workflow against.",
    ),
) -> None:
    """Diff a specific workflow against the installed one."""
    template.diff_template(
        template=workflow, dest=dest, pkg=PKG, template_type=TEMPLATE_TYPE
    )


@CLI.command(name="install")
def install_workflow(
    workflow: str = typer.Argument("all", help="name of the workflow to install."),
    dest: Path = typer.Argument(
        Path("./.github/workflows"), help="target directory to install the workflow to."
    ),
) -> None:
    """
    Installs the requested workflow into the target directory.

    Attention: If there is an existing workflow with the same name it will be overwritten!
    """
    template.install_template(
        template=workflow, dest=dest, pkg=PKG, template_type=TEMPLATE_TYPE
    )


@CLI.command(name="update")
def update_workflow(
    workflow: str = typer.Argument("all", help="name of the workflow to install."),
    dest: Path = typer.Argument(
        Path("./.github/workflows"), help="target directory to install the workflow to."
    ),
    confirm: bool = typer.Option(
        False, help="Automatically confirm overwriting the existing workflow(s)"
    ),
) -> None:
    """Similar to install but checks for existing workflows and shows diff"""
    template.update_template(
        template=workflow,
        dest=dest,
        confirm=confirm,
        pkg=PKG,
        template_type=TEMPLATE_TYPE,
    )


if __name__ == "__main__":
    CLI()



from __future__ import annotations

from typing import Optional

from packaging.version import Version
from pydantic import (
    BaseModel,
    ConfigDict,
)

from exasol.toolbox.util.dependencies.shared_models import (
    NormalizedPackageStr,
    Package,
)


class DependencyChange(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    name: str


class AddedDependency(DependencyChange):
    version: Version

    def __str__(self) -> str:
        return f"* Added dependency `{self.name}:{self.version}`"

    @classmethod
    def from_package(cls, package: Package) -> AddedDependency:
        return cls(name=package.name, version=package.version)


class RemovedDependency(DependencyChange):
    version: Version

    def __str__(self) -> str:
        return f"* Removed dependency `{self.name}:{self.version}`"

    @classmethod
    def from_package(cls, package: Package) -> RemovedDependency:
        return cls(name=package.name, version=package.version)


class UpdatedDependency(DependencyChange):
    previous_version: Version
    current_version: Version

    def __str__(self) -> str:
        return (
            f"* Updated dependency `{self.name}:{self.previous_version}` "
            f"to `{self.current_version}`"
        )

    @classmethod
    def from_package(
        cls, previous_package: Package, current_package: Package
    ) -> UpdatedDependency:
        return cls(
            name=previous_package.name,
            previous_version=previous_package.version,
            current_version=current_package.version,
        )


class DependencyChanges(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    previous_dependencies: dict[NormalizedPackageStr, Package]
    current_dependencies: dict[NormalizedPackageStr, Package]

    def _categorize_change(
        self, dependency_name: NormalizedPackageStr
    ) -> Optional[DependencyChange]:
        """
        Categorize dependency change as removed, added, or updated.
        """
        previous_dependency = self.previous_dependencies.get(dependency_name)
        current_dependency = self.current_dependencies.get(dependency_name)
        if previous_dependency and not current_dependency:
            return RemovedDependency.from_package(previous_dependency)
        elif not previous_dependency and current_dependency:
            return AddedDependency.from_package(current_dependency)
        elif previous_dependency.version != current_dependency.version:  # type: ignore
            return UpdatedDependency.from_package(
                previous_dependency, current_dependency  # type: ignore
            )
        # dependency was unchanged between versions
        return None

    @property
    def changes(self) -> list[DependencyChange]:
        """
        Return dependency changes
        """
        # dict.keys() returns a set converted into a list by `sorted()`
        all_dependencies = sorted(
            self.previous_dependencies.keys() | self.current_dependencies.keys()
        )
        return [
            change_dependency
            for dependency_name in all_dependencies
            if (change_dependency := self._categorize_change(dependency_name))
        ]



import json
import os
import re
import shutil
import sqlite3
import subprocess  # nosec
import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

import nox
from nox import Session

from exasol.toolbox.nox._shared import MINIMUM_PYTHON_VERSION
from noxconfig import (
    PROJECT_CONFIG,
    Config,
)

COVERAGE_DB = ".coverage"
COVERAGE_XML = "ci-coverage.xml"
LINT_JSON = ".lint.json"
LINT_TXT = ".lint.txt"
SECURITY_JSON = ".security.json"

ALL_LINT_FILES = {COVERAGE_DB, LINT_JSON, LINT_TXT, SECURITY_JSON}
COVERAGE_TABLES = {"coverage_schema", "meta", "file", "line_bits"}
LINT_JSON_ATTRIBUTES = {
    "type",
    "module",
    "obj",
    "line",
    "column",
    "endLine",
    "endColumn",
    "path",
    "symbol",
    "message",
    "message-id",
}

SECURITY_JSON_ATTRIBUTES = {"errors", "generated_at", "metrics", "results"}


@nox.session(name="artifacts:validate", python=False)
def check_artifacts(session: Session) -> None:
    """Validate that all project artifacts are available and consistent"""
    all_files = {f.name for f in PROJECT_CONFIG.root.iterdir() if f.is_file()}
    if missing_files := (ALL_LINT_FILES - all_files):
        print(f"files not available: {missing_files}", file=sys.stderr)
        sys.exit(1)

    all_is_valid_checks = [
        _is_valid_lint_txt(Path(PROJECT_CONFIG.root, LINT_TXT)),
        _is_valid_lint_json(Path(PROJECT_CONFIG.root, LINT_JSON)),
        _is_valid_security_json(Path(PROJECT_CONFIG.root, SECURITY_JSON)),
        _is_valid_coverage(Path(PROJECT_CONFIG.root, COVERAGE_DB)),
    ]
    if not all(all_is_valid_checks):
        sys.exit(1)


def _handle_validation_error(file: Path, message: str) -> bool:
    print(f"error in [{file.name}]: {message}", file=sys.stderr)
    return False


def _is_valid_lint_txt(file: Path) -> bool:
    content = file.read_text()
    expr = re.compile(r"^Your code has been rated at (\d+.\d+)/.*", re.MULTILINE)
    matches = expr.search(content)
    if not matches:
        _handle_validation_error(file, "Could not find a rating")
        return False
    return True


def _is_valid_lint_json(file: Path) -> bool:
    try:
        content = file.read_text()
        issues = json.loads(content)
    except json.JSONDecodeError as ex:
        _handle_validation_error(file, f"Invalid json file, details: {ex}")
        return False

    for number, issue in enumerate(issues):
        actual = set(issue.keys())
        missing = LINT_JSON_ATTRIBUTES - actual
        if len(missing) > 0:
            _handle_validation_error(
                file,
                f"Invalid format, issue {number} is missing the following attributes {missing}",
            )
            return False
    return True


def _is_valid_security_json(file: Path) -> bool:
    try:
        content = file.read_text()
        actual = set(json.loads(content))
    except json.JSONDecodeError as ex:
        return _handle_validation_error(file, f"Invalid json file, details: {ex}")

    missing = SECURITY_JSON_ATTRIBUTES - actual
    if len(missing) > 0:
        return _handle_validation_error(
            file,
            f"Invalid format, the file is missing the following attributes {missing}",
        )
    return True


def _is_valid_coverage(path: Path) -> bool:
    try:
        conn = sqlite3.connect(path)
        cursor = conn.cursor()
    except sqlite3.Error as ex:
        return _handle_validation_error(
            path, f"database connection not possible, details: {ex}"
        )
    try:
        actual_tables = set(
            cursor.execute("select name from sqlite_schema where type == 'table'")
        )
    except sqlite3.Error as ex:
        return _handle_validation_error(
            path, f"schema query not possible, details: {ex}"
        )
    actual = {f[0] for f in actual_tables if (f[0] in COVERAGE_TABLES)}
    missing = COVERAGE_TABLES - actual
    if len(missing) > 0:
        return _handle_validation_error(
            path,
            f"Invalid database, the database is missing the following tables {missing}",
        )
    return True


@nox.session(name="artifacts:copy", python=False)
def copy_artifacts(session: Session) -> None:
    """
    Copy artifacts to the current directory
    """

    artifact_dir = Path(session.posargs[0])
    suffix = _python_version_suffix()
    _combine_coverage(session, artifact_dir, f"coverage{suffix}*/{COVERAGE_DB}")
    _copy_artifacts(
        artifact_dir,
        artifact_dir.parent,
        [
            f"lint{suffix}/{LINT_TXT}",
            f"lint{suffix}/{LINT_JSON}",
            f"security{suffix}/{SECURITY_JSON}",
        ],
    )


def _python_version_suffix() -> str:
    versions = getattr(PROJECT_CONFIG, "python_versions", None)
    pivot = versions[0] if versions else MINIMUM_PYTHON_VERSION
    return f"-python{pivot}"


def _combine_coverage(session: Session, artifact_dir: Path, pattern: str):
    """
    pattern: glob pattern, e.g. "*.coverage"
    """
    if args := [f for f in artifact_dir.glob(pattern) if f.exists()]:
        session.run("coverage", "combine", "--keep", *sorted(args))
    else:
        print(f"Could not find any file {artifact_dir}/{pattern}", file=sys.stderr)


def _copy_artifacts(source: Path, dest: Path, files: Iterable[str]):
    for file in files:
        path = source / file
        if path.exists():
            print(f"Copying file {path}", file=sys.stderr)
            shutil.copy(path, dest)
        else:
            print(f"File not found {path}", file=sys.stderr)


def _prepare_coverage_xml(
    session: Session, source: Path, cwd: Optional[Path] = None
) -> None:
    """
    Prepare the coverage XML for input into Sonar

    The coverage XML is used within Sonar to determine the overall coverage in our
    project and the coverage associated with source code changes made in a PR. If the
    code coverage associated with changes made in our PR does not meet the set Sonar
    guidelines for that project, then when the sonarqubecloud bot writes in the PR, it
    will indicate that the coverage constraint was not met, &, if the bot is required in
    the branch protections for that project, then it will block the PR from being
    merged. Thus, in the preparation of the coverage XML, we set `--fail-under=0` so
    that this sub-step will not fail prior to sending the data to Sonar. Otherwise, this
    sub-step would fail whenever the `fail_under` in `[tool.coverage.report]` of the
    `pyproject.toml` was not met.
    """
    command = [
        "coverage",
        "xml",
        "-o",
        COVERAGE_XML,
        "--include",
        f"{source}/*",
        "--fail-under=0",
    ]
    output = subprocess.run(command, capture_output=True, text=True, cwd=cwd)  # nosec
    if output.returncode != 0:
        if output.stdout.strip() == "No data to report.":
            # Assuming that previous steps passed in the CI, this indicates — as
            # is in the case for newly created projects — that no coverage over the
            # `source` files was found. In this scenario, we do not need the CI to
            # break from a non-zero exit code. This is because without a coverage file
            # the sonarqubecloud bot will still report that the coverage is 0%, but as
            # long as the bot is not required in the project's branch protections, then
            # the PR can still be merged, provided that the other required constraints
            # have been met.
            return
        session.error(output.returncode, output.stdout, output.stderr)


def _upload_to_sonar(
    session: Session, sonar_token: Optional[str], config: Config
) -> None:
    command = [
        "pysonar",
        "--sonar-token",
        sonar_token,
        "--sonar-python-pylint-report-path",
        LINT_JSON,
        "--sonar-python-bandit-report-paths",
        SECURITY_JSON,
        "--sonar-python-version",
        ",".join(config.python_versions),
        "--sonar-sources",
        config.source,
    ]
    if Path(COVERAGE_XML).exists():
        command.extend(["--sonar-python-coverage-report-paths", COVERAGE_XML])
    session.run(*command)  # type: ignore


@nox.session(name="sonar:check", python=False)
def upload_artifacts_to_sonar(session: Session) -> None:
    """Upload artifacts to sonar for analysis"""
    sonar_token = os.getenv("SONAR_TOKEN")
    _prepare_coverage_xml(session, PROJECT_CONFIG.source)
    _upload_to_sonar(session, sonar_token, PROJECT_CONFIG)



import json
import logging

import nox
from nox import Session

from noxconfig import (
    PROJECT_CONFIG,
    Config,
)

_log = logging.getLogger(__name__)

_PYTHON_VERSIONS = ["3.9", "3.10", "3.11", "3.12"]
_EXASOL_VERSIONS = ["7.1.9"]


def _python_matrix(config: Config):
    attr = "python_versions"
    python_versions = getattr(config, attr, _PYTHON_VERSIONS)
    if not hasattr(config, attr):
        _log.warning(
            "Config does not contain '%s' setting. Using default: %s",
            attr,
            _PYTHON_VERSIONS,
        )
    return {"python-version": python_versions}


def _exasol_matrix(config: Config):
    attr = "exasol_versions"
    exasol_versions = getattr(config, attr, _EXASOL_VERSIONS)
    if not hasattr(config, attr):
        _log.warning(
            "Config does not contain '%s' setting. Using default: %s",
            attr,
            _EXASOL_VERSIONS,
        )
    return {"exasol-version": exasol_versions}


@nox.session(name="matrix:python", python=False)
def python_matrix(session: Session) -> None:
    """Output the build matrix for Python versions as JSON."""
    print(json.dumps(_python_matrix(PROJECT_CONFIG)))


@nox.session(name="matrix:exasol", python=False)
def exasol_matrix(session: Session) -> None:
    """Output the build matrix for Exasol versions as JSON."""
    print(json.dumps(_exasol_matrix(PROJECT_CONFIG)))


@nox.session(name="matrix:all", python=False)
def full_matrix(session: Session) -> None:
    """Output the full build matrix for Python & Exasol versions as JSON."""
    matrix = _python_matrix(PROJECT_CONFIG)
    matrix.update(_exasol_matrix(PROJECT_CONFIG))
    print(json.dumps(matrix))



from __future__ import annotations

import json
from pathlib import Path

import nox
from nox import Session

from exasol.toolbox.util.dependencies.audit import (
    PipAuditException,
    Vulnerabilities,
)
from exasol.toolbox.util.dependencies.licenses import (
    PackageLicenseReport,
    get_licenses,
)
from exasol.toolbox.util.dependencies.poetry_dependencies import get_dependencies


@nox.session(name="dependency:licenses", python=False)
def dependency_licenses(session: Session) -> None:
    """Return the packages with their licenses"""
    dependencies = get_dependencies(working_directory=Path())
    licenses = get_licenses()
    license_markdown = PackageLicenseReport(
        dependencies=dependencies, licenses=licenses
    )
    print(license_markdown.to_markdown())


@nox.session(name="dependency:audit", python=False)
def audit(session: Session) -> None:
    """Check for known vulnerabilities"""

    try:
        vulnerabilities = Vulnerabilities.load_from_pip_audit(working_directory=Path())
    except PipAuditException as e:
        session.error(e.return_code, e.stdout, e.stderr)

    security_issue_dict = vulnerabilities.security_issue_dict
    print(json.dumps(security_issue_dict, indent=2))



import datetime
import json
import re
import subprocess
import sys
from collections import defaultdict
from dataclasses import (
    asdict,
    dataclass,
)
from enum import (
    Enum,
    auto,
)
from functools import singledispatch
from inspect import cleandoc
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Union,
)


class Rating(Enum):
    """
    A = Excellent
    B = Good
    C = Satisfactory (Ok, could be better though)
    D = Poor (Improvement necessary)
    E = Bad (Need for action)
    F = Broken (Get it fixed!)
    N/A = Rating is not available
    """

    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    NotAvailable = "N/A"

    def __format__(self, format_spec: str) -> str:
        if format_spec == "n":
            return f"{self.value}"
        return str(self)

    @staticmethod
    def from_score(score: float) -> "Rating":
        if 0 <= score < 1:
            return Rating.F
        elif 1 <= score < 3:
            return Rating.E
        elif 3 <= score < 4:
            return Rating.D
        elif 4 <= score < 6:
            return Rating.C
        elif 6 <= score < 8:
            return Rating.B
        elif 8 <= score <= 10:
            return Rating.A
        else:
            raise ValueError(
                "Uncategorized score, score should be in the following interval [0,10]."
            )

    @staticmethod
    def bandit_rating(score: float) -> "Rating":
        score = round(score, 3)
        if score <= 0.2:
            return Rating.F
        elif 0.2 < score <= 1.6:
            return Rating.E
        elif 1.6 < score <= 3:
            return Rating.D
        elif 3 < score <= 4.4:
            return Rating.C
        elif 4.4 < score <= 5.8:
            return Rating.B
        elif 5.8 < score <= 6:
            return Rating.A
        else:
            raise ValueError(
                "Uncategorized score, score should be in the following interval [0,6]."
            )


@dataclass(frozen=True)
class Report:
    commit: str
    date: datetime.datetime
    coverage: float
    maintainability: Rating
    reliability: Rating
    security: Rating
    technical_debt: Rating


def total_coverage(file: Union[str, Path]) -> float:
    with TemporaryDirectory() as tmpdir:
        tmp_dir = Path(tmpdir)
        report = tmp_dir / "coverage.json"
        p = subprocess.run(
            ["coverage", "json", f"--data-file={file}", "-o", f"{report}"],
            capture_output=True,
            check=False,
            encoding="utf-8",
        )
        stdout = p.stdout.strip()

        if p.returncode != 0:
            message = (
                f"The following command"
                f" returned non-zero exit status {p.returncode}:\n"
                f'  {" ".join(p.args)}\n'
                f"{stdout}"
            )
            if (p.returncode == 1) and (stdout == "No data to report."):
                print(f"{message}\nReturning total coverage 100 %.", file=sys.stderr)
                return 100.0
            else:
                raise RuntimeError(message)

        with open(report, encoding="utf-8") as r:
            data = json.load(r)
            total: float = data["totals"]["percent_covered"]

        return total


def _static_code_analysis(file: Union[str, Path]) -> Rating:
    def pylint(f: Union[str, Path]) -> Rating:
        expr = re.compile(r"^Your code has been rated at (\d+.\d+)/.*", re.MULTILINE)
        with open(f, encoding="utf-8") as results:
            data = results.read()

        matches = expr.search(data)
        if matches:
            groups = matches.groups()
        try:
            group = groups[0]
            score = Rating.from_score(float(group))
        except Exception:
            score = Rating.NotAvailable

        return score

    pylint_score = pylint(file)
    return pylint_score


def maintainability(file: Union[str, Path]) -> Rating:
    return _static_code_analysis(file)


def reliability() -> Rating:
    return Rating.NotAvailable


def security(file: Union[str, Path]) -> Rating:
    with open(file) as json_file:
        security_lint = json.load(json_file)
    return Rating.bandit_rating(_bandit_scoring(security_lint["results"]))


def _bandit_scoring(ratings: list[dict[str, Any]]) -> float:
    def char(value: str, default: str = "H") -> str:
        if value in ["HIGH", "MEDIUM", "LOW"]:
            return value[0]
        return default

    weight = {
        "LL": 1 / 18,
        "LM": 1 / 15,
        "LH": 1 / 12,
        "ML": 1 / 9,
        "MM": 1 / 6,
        "MH": 1 / 3,
    }
    exp = 0.0
    for infos in ratings:
        severity = infos["issue_severity"]
        if severity == "HIGH":
            return 0.0
        index = char(severity) + char(infos["issue_confidence"])
        exp += weight[index]
    return 6 * (2**-exp)


def technical_debt() -> Rating:
    return Rating.NotAvailable


def create_report(
    commit: str,
    date: Optional[datetime.datetime] = None,
    coverage_report: Union[str, Path] = ".coverage",
    pylint_report: Union[str, Path] = ".lint.txt",
    bandit_report: Union[str, Path] = ".security.json",
) -> Report:
    return Report(
        commit=commit,
        date=date if date is not None else datetime.datetime.now(),
        coverage=total_coverage(coverage_report),
        maintainability=maintainability(pylint_report),
        reliability=reliability(),
        security=security(bandit_report),
        technical_debt=technical_debt(),
    )


class Format(Enum):
    Text = auto()
    Json = auto()
    Markdown = auto()

    @staticmethod
    def from_string(value: str) -> "Format":
        key = value.lower()
        dispatcher = {
            "json": Format.Json,
            "markdown": Format.Markdown,
            "text": Format.Text,
        }
        try:
            fmt = dispatcher[key]
        except KeyError as ex:
            raise ValueError(f"No known conversion for [{value}]") from ex
        return fmt


@singledispatch
def color(value: Any) -> str:
    return ""


@color.register
def _rating_color(value: Rating) -> str:
    return {
        Rating.A: "brightgreen",
        Rating.B: "green",
        Rating.C: "yellowgreen",
        Rating.D: "yellow",
        Rating.E: "orange",
        Rating.F: "red",
        Rating.NotAvailable: "black",
    }[value]


@color.register(float)
@color.register(int)
def _coverage_color(value: Union[float, int]) -> str:
    if 0 <= value < 20:
        return _rating_color(Rating.F)
    elif 20 <= value < 50:
        return _rating_color(Rating.E)
    elif 50 <= value < 70:
        return _rating_color(Rating.D)
    elif 70 <= value < 80:
        return _rating_color(Rating.C)
    elif 80 <= value < 90:
        return _rating_color(Rating.B)
    elif 90 <= value <= 100:
        return _rating_color(Rating.A)
    else:
        return _rating_color(Rating.NotAvailable)


def _json(report: Report) -> str:
    def identity(obj: Any) -> Any:
        return obj

    transformation: dict[type, Callable[[Any], Any]] = defaultdict(
        lambda: identity,
        {
            Rating: lambda value: f"{value:n}",
            datetime.datetime: lambda value: str(value),
        },
    )
    data = asdict(report)
    normalized = {k: transformation[type(v)](v) for k, v in data.items()}
    return json.dumps(normalized)


def _markdown(report: Report) -> str:
    col1_width = 25
    col2_width = 75

    def _key(name: str) -> str:
        return name.lower().replace(" ", "_")

    data = asdict(report)
    colors = {name: color(data[name]) for name in data}
    entries = {
        "Commit": "{value}",
        "Date": "{value}",
        "Coverage": "![Coverage](https://img.shields.io/badge/-{value:.2f}%25-{color})",
        "Maintainability": "![Maintainability](https://img.shields.io/badge/-{value:n}-{color})",
        "Reliability": "![Reliability](https://img.shields.io/badge/-{value:n}-{color})",
        "Security": "![Security](https://img.shields.io/badge/-{value:n}-{color})",
        "Technical Debt": "![Technical Debt](https://img.shields.io/badge/-{value:n}-{color})",
    }
    row = f"| {{0:<{col1_width}}} | {{1:<{col2_width}}} | "
    rows = (
        row.format(name, entry.format(value=data[_key(name)], color=colors[_key(name)]))
        for name, entry in entries.items()
    )

    return cleandoc(
        """
    {heading}
    {seperator}
    {entries}
    """
    ).format(
        heading=row.format("Category", "Status"),
        seperator=row.format("-" * col1_width, "-" * col2_width),
        entries="\n".join(rows),
    )


def _text(report: Report) -> str:
    def _name(key: str) -> str:
        return f"{key[0].upper()}{key[1:]}:"

    def _value(value: Any) -> str:
        if isinstance(value, (int, float)):
            return f"{value:.2f}%"
        elif isinstance(value, Rating):
            return f"{value:n}"
        return str(value)

    line = "{0:<25}{1:<75}"
    entries = asdict(report)
    return "\n".join(
        [line.format(_name(key), _value(value)) for key, value in entries.items()]
    )


def format_report(report: Report, fmt: Format) -> str:
    dispatcher = {Format.Json: _json, Format.Markdown: _markdown, Format.Text: _text}
    return dispatcher[fmt](report)



from __future__ import annotations

import json
import subprocess  # nosec
import tempfile
from dataclasses import dataclass
from pathlib import Path
from re import search
from typing import (
    Any,
    Union,
)

from pydantic import BaseModel

from exasol.toolbox.util.dependencies.shared_models import Package

PIP_AUDIT_VULNERABILITY_PATTERN = (
    r"^Found \d+ known vulnerabilit\w{1,3} in \d+ package\w?$"
)


@dataclass
class PipAuditException(Exception):
    return_code: int
    stdout: str
    stderr: str

    def __init__(self, subprocess_output: subprocess.CompletedProcess) -> None:
        self.return_code = subprocess_output.returncode
        self.stdout = subprocess_output.stdout
        self.stderr = subprocess_output.stderr


class Vulnerability(Package):
    id: str
    aliases: list[str]
    fix_versions: list[str]
    description: str

    @classmethod
    def from_audit_entry(
        cls, package_name: str, version: str, vuln_entry: dict[str, Any]
    ) -> Vulnerability:
        """
        Create a Vulnerability from a pip-audit vulnerability entry
        """
        return cls(
            name=package_name,
            version=version,
            id=vuln_entry["id"],
            aliases=vuln_entry["aliases"],
            fix_versions=vuln_entry["fix_versions"],
            description=vuln_entry["description"],
        )

    @property
    def security_issue_entry(self) -> dict[str, Union[str, list[str]]]:
        return {
            "name": self.name,
            "version": str(self.version),
            "refs": [self.id] + self.aliases,
            "description": self.description,
        }


def audit_poetry_files(working_directory: Path) -> str:
    """
    Audit the `pyproject.toml` and `poetry.lock` files

    pip-audit evaluates installed packages. This is to provide
    additional security-related information beyond seeing if a given package
    has a known vulnerability. Thus, to audit our `pyproject.toml` and
    `poetry.lock` files without altering a locally sourced poetry environment,
    this function first exports the locked packages to a requirements.txt file.
    Then, pip-audit evaluates the requirements.txt by installing them to a virtualenv
    and then inspecting the dependencies.
    """

    requirements_txt = "requirements.txt"
    output = subprocess.run(
        ["poetry", "export", "--format=requirements.txt"],
        capture_output=True,
        text=True,
        cwd=working_directory,
    )  # nosec
    if output.returncode != 0:
        raise PipAuditException(subprocess_output=output)

    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)
        (tmpdir / requirements_txt).write_text(output.stdout)

        command = ["pip-audit", "-r", requirements_txt, "-f", "json"]
        output = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=tmpdir,
        )  # nosec

    if output.returncode != 0:
        # pip-audit does not distinguish between 1) finding vulnerabilities
        # and 2) other errors performing the pip-audit (i.e. malformed file);
        # they both map to returncode = 1, so we have our own logic to raise errors
        # for the case of 2) and not 1).
        if not search(PIP_AUDIT_VULNERABILITY_PATTERN, output.stderr.strip()):
            raise PipAuditException(subprocess_output=output)
    return output.stdout


class Vulnerabilities(BaseModel):
    vulnerabilities: list[Vulnerability]

    @classmethod
    def load_from_pip_audit(cls, working_directory: Path) -> Vulnerabilities:
        """
        Convert the pip-audit JSON output into a Vulnerabilities model

        The output from pip-audit is a JSON, which as a dictionary looks like:
        >>> audit_dict = {"dependencies": [
        ... {"name": "alabaster", "version": "0.7.16", "vulns": []},
        ... {"name": "cryptography", "version": "43.0.3", "vulns":
        ... [{"id": "GHSA-79v4-65xg-pq4g", "fix_versions": ["44.0.1"],
        ... "aliases": ["CVE-2024-12797"],
        ... "description": "pyca/cryptography\'s wheels..."}, ...]}]}
        """
        audit_json = audit_poetry_files(working_directory)
        audit_dict = json.loads(audit_json)

        vulnerabilities = []
        for entry in audit_dict["dependencies"]:
            for vuln_entry in entry["vulns"]:
                vulnerabilities.append(
                    Vulnerability.from_audit_entry(
                        package_name=entry["name"],
                        version=entry["version"],
                        vuln_entry=vuln_entry,
                    )
                )
        return Vulnerabilities(vulnerabilities=vulnerabilities)

    @property
    def security_issue_dict(self) -> list[dict[str, Union[str, list[str]]]]:
        return [
            vulnerability.security_issue_entry for vulnerability in self.vulnerabilities
        ]



from __future__ import annotations

from typing import (
    Annotated,
    NewType,
)

from packaging.version import Version
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
)

NormalizedPackageStr = NewType("NormalizedPackageStr", str)

VERSION_TYPE = Annotated[str, AfterValidator(lambda v: Version(v))]


def normalize_package_name(package_name: str) -> NormalizedPackageStr:
    return NormalizedPackageStr(package_name.lower().replace("_", "-"))


class Package(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    name: str
    version: VERSION_TYPE

    @property
    def normalized_name(self) -> NormalizedPackageStr:
        return normalize_package_name(self.name)



import collections
import datetime
import json
import logging
import os
import posixpath

from sphinx import config as sphinx_config
from sphinx.locale import _
from sphinx.util import i18n as sphinx_i18n

from exasol.toolbox.util.version import Version as ExasolVersion
from exasol.toolbox.version import VERSION as PLUGIN_VERSION

logger = logging.getLogger(__name__)

DATE_FMT = "%Y-%m-%d %H:%M:%S %z"
DEFAULT_TAG_WHITELIST = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)$"
DEFAULT_BRANCH_WHITELIST = r"^(master|main)"
DEFAULT_REMOTE_WHITELIST = None
DEFAULT_RELEASED_PATTERN = r"^tags/.*$"
DEFAULT_OUTPUTDIR_FORMAT = r"{ref.name}"
DEFAULT_BUILD_TARGETS = {
    "HTML": {
        "builder": "html",
        "downloadable": False,
        "download_format": "",
    },
}
DEFAULT_CLEAN_INTERMEDIATE_FILES_FLAG = True
ARCHIVE_TYPES = ["zip", "tar", "gztar", "bztar", "xztar"]

Version = collections.namedtuple(
    "Version",
    [
        "name",
        "url",
        "version",
        "release",
        "is_released",
        "artefacts",
    ],
)


class VersionInfo:
    def __init__(self, app, context, metadata, current_version_name):
        self.app = app
        self.context = context
        self.metadata = metadata
        self.current_version_name = current_version_name

    def _dict_to_versionobj(self, v):
        return Version(
            name=v["name"],
            url=self.vpathto(v["name"]),
            version=v["version"],
            release=v["release"],
            is_released=v["is_released"],
            artefacts=[
                {"name": name, "url": self.apathto(name, target)}
                for name, target in v["build_targets"].items()
                if self.apathto(name, target) is not None
            ],
        )

    @property
    def tags(self):
        return [
            self._dict_to_versionobj(v)
            for v in self.metadata.values()
            if v["source"] == "tags"
        ]

    @property
    def branches(self):
        return [
            self._dict_to_versionobj(v)
            for v in self.metadata.values()
            if v["source"] != "tags"
        ]

    @property
    def releases(self):
        return [
            self._dict_to_versionobj(v)
            for v in self.metadata.values()
            if v["is_released"]
        ]

    @property
    def in_development(self):
        return [
            self._dict_to_versionobj(v)
            for v in self.metadata.values()
            if not v["is_released"]
        ]

    def __iter__(self):
        yield from self.branches
        yield from sorted(
            self.tags, key=lambda t: ExasolVersion.from_string(t.name), reverse=True
        )

    def __getitem__(self, name):
        v = self.metadata.get(name)
        if v:
            return self._dict_to_versionobj(v)

    def vhasdoc(self, other_version_name):
        if self.current_version_name == other_version_name:
            return True

        other_version = self.metadata[other_version_name]
        return self.context["pagename"] in other_version["docnames"]

    def vpathto(self, other_version_name):
        if self.current_version_name == other_version_name:
            return "{}.html".format(posixpath.split(self.context["pagename"])[-1])

        # Find relative outputdir paths from common output root
        current_version = self.metadata[self.current_version_name]
        other_version = self.metadata[other_version_name]

        current_outputroot = os.path.abspath(current_version["outputdir"])
        other_outputroot = os.path.abspath(other_version["outputdir"])
        outputroot = os.path.commonpath((current_outputroot, other_outputroot))

        current_outputroot = os.path.relpath(current_outputroot, start=outputroot)
        other_outputroot = os.path.relpath(other_outputroot, start=outputroot)

        # Ensure that we use POSIX separators in the path (for the HTML code)
        if os.sep != posixpath.sep:
            current_outputroot = posixpath.join(*os.path.split(current_outputroot))
            other_outputroot = posixpath.join(*os.path.split(other_outputroot))

        # Find relative path to root of other_version's outputdir
        current_outputdir = posixpath.dirname(
            posixpath.join(current_outputroot, self.context["pagename"])
        )
        other_outputdir = posixpath.relpath(other_outputroot, start=current_outputdir)

        if not self.vhasdoc(other_version_name):
            return posixpath.join(other_outputdir, "index.html")

        return posixpath.join(
            other_outputdir, "{}.html".format(self.context["pagename"])
        )

    def apathto(self, build_target_name, build_target):
        """Find the path to the artefact identified by build_target_name
        and build_target.
        """
        current_version = self.metadata[self.current_version_name]
        current_outputroot = os.path.abspath(current_version["outputdir"])
        artefact_dir = posixpath.join(current_outputroot, "artefacts")
        current_outputdir = posixpath.dirname(
            posixpath.join(current_outputroot, self.context["pagename"])
        )

        filename = "{project}_docs-{version}".format(
            project=self.app.config.project.replace(" ", ""),
            version=self.current_version_name.replace("/", "-"),
        )

        if build_target["download_format"] in ARCHIVE_TYPES:
            filename = "{f}-{build_name}.{extension}".format(
                f=filename,
                build_name=build_target_name,
                extension=build_target["download_format"],
            )
        else:
            filename = "{f}.{extension}".format(
                f=filename,
                extension=build_target["download_format"],
            )
        artefact_path = posixpath.relpath(
            posixpath.join(artefact_dir, filename), start=current_outputdir
        )
        return artefact_path


def html_page_context(app, pagename, templatename, context, doctree):
    versioninfo = VersionInfo(
        app, context, app.config.smv_metadata, app.config.smv_current_version
    )
    context["versions"] = versioninfo
    context["vhasdoc"] = versioninfo.vhasdoc
    context["vpathto"] = versioninfo.vpathto

    current = versioninfo[app.config.smv_current_version]
    context["current_version"] = current.name
    context["latest_version"] = versioninfo[app.config.smv_latest_version]
    context["html_theme"] = app.config.html_theme


def config_inited(app, config):
    """Update the Sphinx builder.
    :param sphinx.application.Sphinx app: Sphinx application object.
    """

    if not config.smv_metadata:
        if not config.smv_metadata_path:
            return

        with open(config.smv_metadata_path, encoding="utf-8") as f:
            metadata = json.load(f)

        config.smv_metadata = metadata

    if not config.smv_current_version:
        return

    try:
        data = app.config.smv_metadata[config.smv_current_version]
    except KeyError:
        return

    app.connect("html-page-context", html_page_context)

    # Restore config values
    old_config = sphinx_config.Config.read(data["confdir"])
    old_config.pre_init_values()
    old_config.init_values()
    config.version = data["version"]
    config.release = data["release"]
    config.rst_prolog = data["rst_prolog"]
    config.today = old_config.today
    if not config.today:
        config.today = sphinx_i18n.format_date(
            format=config.today_fmt or _("%b %d, %Y"),
            date=datetime.datetime.strptime(data["creatordate"], DATE_FMT),
            language=config.language,
        )


def setup(app):
    app.add_config_value("smv_metadata", {}, "html")
    app.add_config_value("smv_metadata_path", "", "html")
    app.add_config_value("smv_current_version", "", "html")
    app.add_config_value("smv_latest_version", "main", "html")
    app.add_config_value("smv_tag_whitelist", DEFAULT_TAG_WHITELIST, "html")
    app.add_config_value("smv_branch_whitelist", DEFAULT_BRANCH_WHITELIST, "html")
    app.add_config_value("smv_remote_whitelist", DEFAULT_REMOTE_WHITELIST, "html")
    app.add_config_value("smv_released_pattern", DEFAULT_RELEASED_PATTERN, "html")
    app.add_config_value("smv_outputdir_format", DEFAULT_OUTPUTDIR_FORMAT, "html")
    app.add_config_value("smv_build_targets", DEFAULT_BUILD_TARGETS, "html")
    app.add_config_value(
        "smv_clean_intermediate_files",
        DEFAULT_CLEAN_INTERMEDIATE_FILES_FLAG,
        "html",
    )
    app.connect("config-inited", config_inited)

    return {
        "version": PLUGIN_VERSION,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from __future__ import annotations

import argparse
from collections import ChainMap
from collections.abc import (
    Iterable,
    MutableMapping,
)
from enum import (
    Enum,
    auto,
)
from pathlib import Path
from typing import Any

from nox import Session

from noxconfig import PROJECT_CONFIG

DEFAULT_PATH_FILTERS = {"dist", ".eggs", "venv", ".poetry"}
DOCS_OUTPUT_DIR = ".html-documentation"

MINIMUM_PYTHON_VERSION = "3.9"


class Mode(Enum):
    Fix = auto()
    Check = auto()


def python_files(project_root: Path) -> Iterable[str]:
    """
    Returns iterable of python files after removing unwanted paths
    """
    deny_list = DEFAULT_PATH_FILTERS.union(set(PROJECT_CONFIG.path_filters))

    files = project_root.glob("**/*.py")
    return [f"{path}" for path in files if not set(path.parts).intersection(deny_list)]


def _version(session: Session, mode: Mode) -> None:
    command = ["nox", "-s", "version:check", "--"]
    command = command if mode == Mode.Check else command + ["--fix"]
    session.run(*command)


def _context_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--db-version", default="7.1.9", help="Specify the Exasol DB version to be used"
    )
    parser.add_argument(
        "--coverage", action="store_true", help="Enable the collection of coverage data"
    )
    return parser


def _context(session: Session, **kwargs: Any) -> MutableMapping[str, Any]:
    parser = _context_parser()
    namespace, args = parser.parse_known_args(session.posargs)
    cli_context: MutableMapping[str, Any] = vars(namespace)
    cli_context["fwd-args"] = args
    default_context = {"db_version": "7.1.9", "coverage": False}
    # Note: ChainMap scans last to first
    return ChainMap(kwargs, cli_context, default_context)



from __future__ import annotations

from datetime import datetime
from inspect import cleandoc
from pathlib import Path

from exasol.toolbox.util.dependencies.poetry_dependencies import (
    get_dependencies,
    get_dependencies_from_latest_tag,
)
from exasol.toolbox.util.dependencies.track_changes import DependencyChanges
from exasol.toolbox.util.version import Version

UNRELEASED_INITIAL_CONTENT = "# Unreleased\n"


class Changelogs:
    def __init__(self, changes_path: Path, root_path: Path, version: Version) -> None:
        """
        Args:
            changes_path: directory containing the changelog & changes files, e.g. `doc/changes/`
            root_path: root directory of the current project, containing file
                `pyproject.toml`
            version: the version to be used in the versioned changes file and listed in
                the `changelog.md`, which contains the index of the change log
        """

        self.version = version
        self.unreleased_md: Path = changes_path / "unreleased.md"
        self.versioned_changelog_md: Path = changes_path / f"changes_{version}.md"
        self.changelog_md: Path = changes_path / "changelog.md"
        self.root_path: Path = root_path

    def _create_new_unreleased(self):
        """
        Write a new unreleased changelog file.
        """
        self.unreleased_md.write_text(UNRELEASED_INITIAL_CONTENT)

    def _create_versioned_changelog(self, unreleased_content: str) -> None:
        """
        Create a versioned changes file.

        Args:
            unreleased_content: the content of the (not yet versioned) changes
        """
        header = f"# {self.version} - {datetime.today().strftime('%Y-%m-%d')}"

        dependency_content = ""
        if dependency_changes := self._describe_dependency_changes():
            dependency_content = f"## Dependency Updates\n{dependency_changes}"

        template = cleandoc(f"{header}\n{unreleased_content}\n{dependency_content}")
        self.versioned_changelog_md.write_text(template + "\n")

    def _extract_unreleased_notes(self) -> str:
        """
        Extract (not yet versioned) changes from `unreleased.md`.
        """
        with self.unreleased_md.open(mode="r", encoding="utf-8") as f:
            # skip header when reading in file, as contains # Unreleased
            lines = f.readlines()[1:]
        unreleased_content = cleandoc("".join(lines))
        return unreleased_content + "\n"

    def _describe_dependency_changes(self) -> str:
        """
        Describe the dependency changes between the latest tag and the current version
        for use in the versioned changes file.
        """
        previous_dependencies_in_groups = get_dependencies_from_latest_tag()
        current_dependencies_in_groups = get_dependencies(
            working_directory=self.root_path
        )

        changes_by_group: list[str] = []
        # dict.keys() returns a set
        all_groups = (
            previous_dependencies_in_groups.keys()
            | current_dependencies_in_groups.keys()
        )
        for group in self._sort_groups(all_groups):
            previous_dependencies = previous_dependencies_in_groups.get(group, {})
            current_dependencies = current_dependencies_in_groups.get(group, {})
            changes = DependencyChanges(
                previous_dependencies=previous_dependencies,
                current_dependencies=current_dependencies,
            ).changes
            if changes:
                changes_str = "\n".join(str(change) for change in changes)
                changes_by_group.append(f"\n### `{group}`\n{changes_str}\n")
        return "".join(changes_by_group)

    @staticmethod
    def _sort_groups(groups: set[str]) -> list[str]:
        """
        Prepare a deterministic sorting for groups shown in the versioned changes file:
            - `main` group should always be first
            - remaining groups are sorted alphabetically
        """
        main = "main"
        if main not in groups:
            # sorted converts set to list
            return sorted(groups)
        remaining_groups = groups - {main}
        # sorted converts set to list
        return [main] + sorted(remaining_groups)

    def _update_changelog_table_of_contents(self) -> None:
        """
        Read in existing `changelog.md` and append to appropriate sections
        before writing out to again.
        """
        updated_content = []
        with self.changelog_md.open(mode="r", encoding="utf-8") as f:
            for line in f:
                updated_content.append(line)
                if line.startswith("* [unreleased]"):
                    updated_content.append(
                        f"* [{self.version}](changes_{self.version}.md)\n"
                    )
                if line.startswith("unreleased"):
                    updated_content.append(f"changes_{self.version}\n")
        updated_content_str = "".join(updated_content)

        self.changelog_md.write_text(updated_content_str)

    def get_changed_files(self) -> list[Path]:
        return [self.unreleased_md, self.versioned_changelog_md, self.changelog_md]

    def update_changelogs_for_release(self) -> None:
        """
        Rotates the changelogs as is needed for a release.

          1. Moves the contents from the `unreleased.md` to the `changes_<version>.md`
          2. Create a new file `unreleased.md`
          3. Updates the table of contents in the `changelog.md` with the new `changes_<version>.md`
        """

        # create versioned changelog
        unreleased_content = self._extract_unreleased_notes()
        self._create_versioned_changelog(unreleased_content)

        # update other changelogs now that versioned changelog exists
        self._create_new_unreleased()
        self._update_changelog_table_of_contents()



import pluggy

_PLUGIN_MARKER = "python-toolbox-nox"
hookspec = pluggy.HookspecMarker("python-toolbox-nox")
hookimpl = pluggy.HookimplMarker("python-toolbox-nox")


class NoxTasks:
    """
    Defines extension hooks for the standard Nox tasks provided by the python-toolbox.

    This class provides an interface that allows various actions to be taken
    when certain events occur within the Nox task execution environment.
    Examples of such events are: the start of a task or the end of a task.

    In Plugins, exceptions should not be thrown. Instead, use the `error` method on the session
    object to abort task execution. However, keep in mind that forcefully stopping execution
    could potentially impact any other code that depends on the execution of the aborted task.
    """

    @hookspec
    def prepare_release_update_version(self, session, config, version):
        """
        This hook is called during version updating when a release is being prepared.
        Implementors can add their own logic and tasks required to be run during the version update here.

        Args:
            session (nox.Session):
                The nox session running the release preparation.
                This it can be used to run commands, etc.

            config (object):
                The project configuration object from the noxconfig.py file.

            version (str):
                A string representation of the version to be released.
                This follows the pattern of Semantic Versioning, i.e., "MAJOR.MINOR.PATCH".
                An example would be "1.4.2".
        """

    @hookspec
    def prepare_release_add_files(self, session, config):
        """
        Files which should be added to the prepare release commit should be added using add.

        Args:
            session (nox.Session):
                The nox session running the release preparation.
                This it can be used to run commands, etc.

            config (object):
                The project configuration object from the noxconfig.py file.

        Return:
            A iterable of Path objects of files to be added.

            e.g.:  [Path("file1.txt")] or [Path("file1.txt"), Path("file2.txt")]
        """

    @hookspec
    def pre_integration_tests_hook(self, session, config, context):
        """
        Implement if project specific behaviour is required before running integration tests.

        This function acts as a hook that gets called before the execution of integration tests.
        It can be used to execute any project-specific tasks that need to be performed before
        the tests run, such as starting or initializing a database.

        Args:
            session (nox.Session):
                The nox session that ran the integration tests.
                It can be used to run additional commands, etc.

            config (object):
                The project's configuration object from the noxconfig.py file.

            context (dict):
                Contains additional contextual information related to the testing session.
                All keys in this dictionary will be strings, while the values can be of any type.
        """

    @hookspec
    def post_integration_tests_hook(self, session, config, context):
        """
        Implement if project specific behaviour is required after running integration tests.

        This function acts as a hook that gets called after the execution of integration tests.
        It can be used to execute any project-specific tasks that need to be performed after
        the tests run, such as clean up, data processing, reporting etc.

        Args:
            session (nox.Session):
                The nox session that ran the integration tests.
                It can be used to run additional commands, etc.

            config (object):
                The project's configuration object from the noxconfig.py file.

            context (dict):
                Contains additional contextual information related to the testing session.
                All keys in this dictionary will be strings, while the values can be of any type.
        """

    @staticmethod
    def plugin_manager(config) -> pluggy.PluginManager:
        pm = pluggy.PluginManager(_PLUGIN_MARKER)
        pm.add_hookspecs(NoxTasks)
        for plugin in getattr(config, "plugins", []):
            pm.register(plugin())
        return pm



from argparse import ArgumentTypeError

from exasol.toolbox.util.version import Version


def version(arg: str) -> Version:
    try:
        return Version.from_string(arg)
    except Exception as ex:
        msg = f"Expected format: <number>.<number>.<number>, e.g. 1.2.3, actual: {arg}"
        raise ArgumentTypeError(msg) from ex



from __future__ import annotations

import subprocess
import tempfile
from collections import OrderedDict
from dataclasses import dataclass
from inspect import cleandoc
from json import loads
from typing import Optional

from pydantic import field_validator

from exasol.toolbox.util.dependencies.shared_models import (
    NormalizedPackageStr,
    Package,
)

LICENSE_MAPPING_TO_ABBREVIATION = {
    "BSD License": "BSD",
    "MIT License": "MIT",
    "The Unlicensed (Unlicensed)": "Unlicensed",
    "Mozilla Public License 2.0 (MPL 2.0)": "MPLv2",
    "GNU General Public License (GPL)": "GPL",
    "GNU Lesser General Public License v2 (LGPLv2)": "LGPLv2",
    "GNU General Public License v2 (GPLv2)": "GPLv2",
    "GNU General Public License v2 or later (GPLv2+)": "GPLv2+",
    "GNU General Public License v3 (GPLv3)": "GPLv3",
    "Apache Software License": "Apache",
}

LICENSE_MAPPING_TO_URL = {
    "GPLv1": "https://www.gnu.org/licenses/old-licenses/gpl-1.0.html",
    "GPLv2": "https://www.gnu.org/licenses/old-licenses/gpl-2.0.html",
    "LGPLv2": "https://www.gnu.org/licenses/old-licenses/lgpl-2.0.html",
    "GPLv3": "https://www.gnu.org/licenses/gpl-3.0.html",
    "LGPLv3": "https://www.gnu.org/licenses/lgpl-3.0.html",
    "Apache": "https://www.apache.org/licenses/LICENSE-2.0",
    "MIT": "https://mit-license.org/",
    "BSD": "https://opensource.org/license/bsd-3-clause",
}


class PackageLicense(Package):
    package_link: Optional[str]
    license: str

    @field_validator("package_link", mode="before")
    def map_unknown_to_none(cls, v) -> Optional[str]:
        if v == "UNKNOWN":
            return None
        return v

    @field_validator("license", mode="before")
    def map_to_normalized_values(cls, v) -> Optional[str]:
        return _normalize(v)

    @property
    def license_link(self) -> Optional[str]:
        return LICENSE_MAPPING_TO_URL.get(self.license, None)


def _normalize(_license: str) -> str:
    def is_multi_license(l: str) -> bool:
        return ";" in l

    def select_most_restrictive(licenses: list[str]) -> str:
        lic = "Unknown"
        _mapping = {
            "Unknown": -1,
            "Unlicensed": 0,
            "BSD": 1,
            "MIT": 2,
            "MPLv2": 3,
            "LGPLv2": 4,
            "GPLv2": 5,
            "GPLv3": 6,
        }
        for l in licenses:
            if l in _mapping:
                if _mapping[l] > _mapping[lic]:
                    lic = l
            else:
                return "<br>".join(licenses)
        return lic

    if is_multi_license(_license):
        items = []
        for item in _license.split(";"):
            item = str(item).strip()
            items.append(LICENSE_MAPPING_TO_ABBREVIATION.get(item, item))
        return select_most_restrictive(items)

    return LICENSE_MAPPING_TO_ABBREVIATION.get(_license, _license)


def _packages_from_json(json: str) -> dict[NormalizedPackageStr, PackageLicense]:
    packages = loads(json)
    return {
        package_license.normalized_name: package_license
        for package in packages
        if (
            package_license := PackageLicense(
                name=package["Name"],
                package_link=package["URL"],
                version=package["Version"],
                license=package["License"],
            )
        )
    }


def get_licenses() -> dict[NormalizedPackageStr, PackageLicense]:
    with tempfile.NamedTemporaryFile() as file:
        subprocess.run(
            [
                "pip-licenses",
                "--format=json",
                "--output-file=" + file.name,
                "--with-system",
                "--with-urls",
            ],
            capture_output=True,
            check=True,
        )
        return _packages_from_json(file.read().decode())


@dataclass(frozen=True)
class PackageLicenseReport:
    dependencies: OrderedDict[str, dict[NormalizedPackageStr, Package]]
    licenses: dict[NormalizedPackageStr, PackageLicense]

    @staticmethod
    def _format_group_table_header(group: str) -> str:
        return cleandoc(
            f"""## `{group}` Dependencies
            |Package|Version|License|
            |---|---|---|
            """
        )

    def _format_group_table(
        self, group: str, group_package_names: set[NormalizedPackageStr]
    ) -> str:
        group_header = self._format_group_table_header(group=group)

        rows = []
        for package_name in sorted(group_package_names):
            if license_info := self.licenses.get(package_name):
                rows.append(self._format_table_row(license_info=license_info))

        return f"""{group_header}\n{''.join(rows)}\n"""

    @staticmethod
    def _format_table_row(license_info: PackageLicense) -> str:
        row_package = f"{license_info.name}"
        if license_info.package_link:
            row_package = f"[{license_info.name}]({license_info.package_link})"

        row_license = f"{license_info.license}"
        if license_info.license_link:
            row_license = f"[{license_info.license}]({license_info.license_link})"

        return f"|{row_package}|{license_info.version}|{row_license}|\n"

    def to_markdown(self) -> str:
        rows = []
        for group in self.dependencies:
            group_package_names = set(self.dependencies[group].keys())
            rows.append(
                self._format_group_table(
                    group=group, group_package_names=group_package_names
                )
            )
        return cleandoc(f"""# Dependencies\n\n{''.join(rows)}""")



from __future__ import annotations

import subprocess
import tempfile
from collections import OrderedDict
from pathlib import Path
from typing import Optional

import tomlkit
from pydantic import (
    BaseModel,
    ConfigDict,
)
from tomlkit import TOMLDocument

from exasol.toolbox.util.dependencies.shared_models import (
    NormalizedPackageStr,
    Package,
)
from exasol.toolbox.util.git import Git


class PoetryGroup(BaseModel):
    model_config = ConfigDict(frozen=True)

    name: str
    toml_section: Optional[str]


PYPROJECT_TOML = "pyproject.toml"
TRANSITIVE_GROUP = PoetryGroup(name="transitive", toml_section=None)


class PoetryToml(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    content: TOMLDocument

    @classmethod
    def load_from_toml(cls, working_directory: Path) -> PoetryToml:
        file_path = working_directory / PYPROJECT_TOML
        if not file_path.exists():
            raise ValueError(f"File not found: {file_path}")

        try:
            text = file_path.read_text()
            content = tomlkit.loads(text)
            return cls(content=content)
        except Exception as e:
            raise ValueError(f"Error reading file: {str(e)}")

    def get_section_dict(self, section: str) -> Optional[dict]:
        current = self.content.copy()
        for section in section.split("."):
            if section not in current:
                return None
            current = current[section]  # type: ignore
        return current

    @property
    def groups(self) -> tuple[PoetryGroup, ...]:
        groups = []

        main_key = "project.dependencies"
        if self.get_section_dict(main_key):
            groups.append(PoetryGroup(name="main", toml_section=main_key))

        main_dynamic_key = "tool.poetry.dependencies"
        if self.get_section_dict(main_dynamic_key):
            groups.append(PoetryGroup(name="main", toml_section=main_dynamic_key))

        group_key = "tool.poetry.group"
        if group_dict := self.get_section_dict(group_key):
            for group, content in group_dict.items():
                if "dependencies" in content:
                    groups.append(
                        PoetryGroup(
                            name=group,
                            toml_section=f"{group_key}.{group}.dependencies",
                        )
                    )
        return tuple(groups)


class PoetryDependencies(BaseModel):
    groups: tuple[PoetryGroup, ...]
    working_directory: Path

    @staticmethod
    def _extract_from_line(line: str) -> Optional[Package]:
        # remove (!) from line as indicates not installed in environment,
        # which could occur for optional dependencies
        split_line = line.replace("(!)", "").strip().split(maxsplit=2)
        if len(split_line) < 2:
            print(f"Unable to parse dependency={line}")
            return None
        return Package(name=split_line[0], version=split_line[1])

    def _extract_from_poetry_show(
        self, output_text: str
    ) -> dict[NormalizedPackageStr, Package]:
        return {
            package.normalized_name: package
            for line in output_text.splitlines()
            if (package := self._extract_from_line(line))
        }

    @property
    def direct_dependencies(
        self,
    ) -> OrderedDict[str, dict[NormalizedPackageStr, Package]]:
        dependencies = OrderedDict()
        for group in self.groups:
            command = (
                "poetry",
                "show",
                "--top-level",
                f"--only={group.name}",
                "--no-truncate",
            )
            output = subprocess.run(
                command,
                capture_output=True,
                text=True,
                cwd=self.working_directory,
                check=True,
            )
            result = self._extract_from_poetry_show(output_text=output.stdout)
            dependencies[group.name] = result
        return dependencies

    @property
    def all_dependencies(self) -> OrderedDict[str, dict[NormalizedPackageStr, Package]]:
        command = ("poetry", "show", "--no-truncate")
        output = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=self.working_directory,
            check=True,
        )

        direct_dependencies = self.direct_dependencies.copy()

        transitive_dependencies = {}
        names_direct_dependencies = {
            package_name
            for group_list in direct_dependencies
            for package_name in group_list
        }
        for line in output.stdout.splitlines():
            dep = self._extract_from_line(line=line)
            if dep and dep.name not in names_direct_dependencies:
                transitive_dependencies[dep.normalized_name] = dep

        return direct_dependencies | {TRANSITIVE_GROUP.name: transitive_dependencies}


def get_dependencies(
    working_directory: Path,
) -> OrderedDict[str, dict[NormalizedPackageStr, Package]]:
    poetry_dep = PoetryToml.load_from_toml(working_directory=working_directory)
    return PoetryDependencies(
        groups=poetry_dep.groups, working_directory=working_directory
    ).direct_dependencies


def get_dependencies_from_latest_tag() -> (
    OrderedDict[str, dict[NormalizedPackageStr, Package]]
):
    latest_tag = Git.get_latest_tag()
    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)

        Git.copy_remote_file_locally(latest_tag, "poetry.lock", tmpdir)
        Git.copy_remote_file_locally(latest_tag, PYPROJECT_TOML, tmpdir)

        return get_dependencies(working_directory=tmpdir)



import difflib
import io
from collections.abc import Mapping
from contextlib import ExitStack
from pathlib import Path
from typing import (
    Any,
    Union,
)

import importlib_resources as resources
import typer
from rich.columns import Columns
from rich.console import Console
from rich.syntax import Syntax

stdout = Console()
stderr = Console(stderr=True)

CLI = typer.Typer()


def _templates(pkg: str) -> Mapping[str, Any]:
    def _normalize(name: str) -> str:
        name, _ = name.split(".")
        return name

    return {_normalize(w.name): w for w in resources.files(pkg).iterdir()}


def list_templates(columns: bool, pkg: str) -> None:
    """List all available templates."""

    class List:
        def __init__(self, items: Any):
            self.items = items

        def __rich__(self) -> str:
            return "\n".join(self.items)

    output = (
        List(_templates(pkg)) if not columns else Columns(_templates(pkg), expand=True)
    )
    stdout.print(output)


def show_templates(
    template: str,
    pkg: str,
    template_type: str,
    lexer: str,
) -> None:
    """Shows a specific template."""
    templates = _templates(pkg)
    if template not in templates:
        stdout.print(f"Unknown {template_type} <{template}>.", style="red")
        raise typer.Exit(code=1)

    template = templates[template]
    stdout.print(Syntax.from_path(path=template, encoding="utf-8", lexer=lexer))  # type: ignore


def diff_template(template: str, dest: Path, pkg: str, template_type: str) -> None:
    """Diff a specific template against the installed one."""
    templates = _templates(pkg)
    if template not in templates:
        stdout.print(f"Unknown {template_type} <{template}>.", style="red")
        raise typer.Exit(code=1)

    # Use Any type to enable reuse of the variable/binding name
    for name, path in templates.items():
        if name == template:
            old: Any = dest / f"{template}{path.suffix}"
            new: Any = Path(_templates(pkg)[template])
            with ExitStack() as stack:
                old = stack.enter_context(
                    open(old, encoding="utf-8") if old.exists() else io.StringIO("")
                )
                new = stack.enter_context(open(new, encoding="utf-8"))
                old = old.read().split("\n")
                new = new.read().split("\n")

            diff = difflib.unified_diff(old, new, fromfile="old", tofile="new")
            stdout.print(Syntax("\n".join(diff), "diff"))


def _install_template(
    template_type: str,
    src: Union[str, Path],
    dest: Union[str, Path],
    exists_ok: bool = False,
) -> None:
    src, dest = Path(src), Path(dest)

    if dest.exists() and not exists_ok:
        raise FileExistsError(f"{template_type} already exists")

    with ExitStack() as stack:
        input_file = stack.enter_context(open(src, "rb"))
        output_file = stack.enter_context(open(dest, "wb"))
        output_file.write(input_file.read())


def _select_templates(template: str, pkg: str, template_type: str) -> Mapping[str, Any]:
    templates = _templates(pkg)
    if template != "all" and template not in templates:
        raise Exception(f"{template_type} <{template}> is unknown")
    templates = (
        templates
        if template == "all"
        else {name: path for name, path in templates.items() if template == name}
    )
    return templates


def install_template(template: str, dest: Path, pkg: str, template_type: str) -> None:
    """
    Installs the requested template into the target directory.

    Attention: If there is an existing template with the same name it will be overwritten!
    """
    if not dest.exists():
        dest.mkdir(parents=True)

    try:
        templates = _select_templates(template, pkg, template_type)
    except Exception as ex:
        stderr.print(f"[red]{ex}[/red]")
        raise typer.Exit(-1)

    for name, path in templates.items():
        destination = dest / f"{name}{path.suffix}"
        _install_template(template_type, path, destination, exists_ok=True)
        stderr.print(f"Installed {name} in {destination}")


def update_template(
    template: str, dest: Path, confirm: bool, pkg: str, template_type: str
) -> None:
    """Similar to install but checks for existing templates and shows diff"""
    if not dest.exists():
        dest.mkdir(parents=True)

    try:
        templates = _select_templates(template, pkg, template_type)
    except Exception as ex:
        stderr.print(f"[red]{ex}[/red]")
        raise typer.Exit(-1)

    if confirm:
        install_template(template, dest, pkg, template_type)
        raise typer.Exit(0)

    for name, path in templates.items():
        destination = dest / f"{name}{path.suffix}"
        try:
            _install_template(template_type, path, destination, exists_ok=False)
            stderr.print(f"Updated {name} in {destination}")
        except FileExistsError:
            show_diff = typer.confirm(
                f"{template_type} <{name}> already exists, show diff?"
            )
            if show_diff:
                diff_template(name, dest, pkg, template_type)

            overwrite = typer.confirm(f"Overwrite existing {template_type}?")
            if overwrite:
                _install_template(template_type, path, destination, exists_ok=True)
                stderr.print(f"Updated {name} in {destination}")


if __name__ == "__main__":
    CLI()
